import datetime
from pathlib import Path
from typing import Iterable, Optional, Generic, Union, Callable, NamedTuple, Tuple, List

import jinja2
from markupsafe import Markup

from ..log.generic import PersistenceVariable, LoggingRawWebUIView, AggregationMethod, LoggingAggregatedWebUIView
from ..base import T
from .interface import WebPageItem, WebUIConnector, jinja_env


class LogListDataSpec(NamedTuple):
    variable: PersistenceVariable
    format: Union[str, Markup, Callable[[Union[T, float]], Union[str, Markup]]] = "{}"
    color: str = ''
    aggregation: Optional[AggregationMethod] = None
    aggregation_interval: Optional[datetime.timedelta] = None


class LogListWidget(WebPageItem, Generic[T]):
    def __init__(self, interval: datetime.timedelta, data_spec: List[LogListDataSpec]):
        self.interval = interval

        self.specs = []
        self.connectors = []

        for spec in data_spec:
            formatter: Callable[[T], Union[str, Markup]] = (
                (lambda x: spec.format.format(x))  # type: ignore
                if isinstance(spec.format, (str, Markup))
                else spec.format)
            aggregation_interval = spec.aggregation_interval if spec.aggregation_interval is not None else interval / 10
            connector = (LoggingRawWebUIView(spec.variable, interval, formatter, include_previous=False)
                         if spec.aggregation is None
                         else LoggingAggregatedWebUIView(spec.variable, interval, spec.aggregation,
                                                         aggregation_interval, formatter))
            self.connectors.append(connector)
            self.specs.append({'id': id(connector),
                               'color': spec.color})

    async def render(self) -> str:
        return await jinja_env.get_template('log/loglist.htm').render_async(
            spec=self.specs, interval=round(self.interval.total_seconds() * 1000))

    def get_connectors(self) -> Iterable[WebUIConnector]:
        return self.connectors


class ChartDataSpec(NamedTuple):
    variable: PersistenceVariable
    label: str
    color: Optional[Tuple[int, int, int]] = None
    aggregation: Optional[AggregationMethod] = None
    aggregation_interval: Optional[datetime.timedelta] = None


class ChartWidget(WebPageItem):
    COLORS = [
        (0, 181, 173),
        (219, 40, 40),
        (33, 186, 69),
        (163, 51, 200),
        (251, 189, 8),
    ]

    def __init__(self, interval: datetime.timedelta,  data_spec: List[ChartDataSpec]):
        self.interval = interval
        self.align_ticks_to = datetime.datetime(2020, 1, 1, 0, 0, 0)
        self.row_specs = []
        self.connectors = []

        for i, spec in enumerate(data_spec):
            aggregation_interval = spec.aggregation_interval
            if aggregation_interval is None:
                aggregation_interval = interval / 10
            is_aggregated = spec.aggregation is not None
            connector = (LoggingRawWebUIView(spec.variable, interval, include_previous=True)
                         if spec.aggregation is None
                         else LoggingAggregatedWebUIView(spec.variable, interval, spec.aggregation,
                                                         aggregation_interval, align_to=self.align_ticks_to))
            self.connectors.append(connector)
            self.row_specs.append({'id': id(connector),
                                   'is_aggregated': is_aggregated,
                                   'color': spec.color if spec.color is not None else self.COLORS[i % len(self.COLORS)],
                                   'label': spec.label})

    async def render(self) -> str:
        return await jinja_env.get_template('log/chart.htm').render_async(
            spec=self.row_specs, interval=round(self.interval.total_seconds() * 1000),
            align_ticks_to=self.align_ticks_to.timestamp() * 1000)

    def get_connectors(self) -> Iterable[WebUIConnector]:
        return self.connectors




import unittest

from shc import datatypes, conversion


class ConversionTests(unittest.TestCase):
    def test_ranges_1(self) -> None:
        v0 = 0.5
        conv0 = conversion.get_converter(float, datatypes.RangeFloat1)
        r1 = conv0(v0)
        self.assertIsInstance(r1, datatypes.RangeFloat1)

        conv1 = conversion.get_converter(datatypes.RangeFloat1, float)
        v1 = conv1(r1)
        self.assertEqual(0.5, v1)

        conv2 = conversion.get_converter(datatypes.RangeFloat1, datatypes.RangeUInt8)
        r2 = conv2(r1)
        self.assertIsInstance(r2, datatypes.RangeUInt8)
        conv3 = conversion.get_converter(datatypes.RangeUInt8, int)
        v2 = conv3(r2)
        self.assertEqual(128, v2)

        conv4 = conversion.get_converter(datatypes.RangeUInt8, datatypes.RangeInt0To100)
        r3 = conv4(r2)
        self.assertIsInstance(r3, datatypes.RangeInt0To100)
        conv5 = conversion.get_converter(datatypes.RangeInt0To100, int)
        v3 = conv5(r3)
        self.assertEqual(50, v3)

        conv6 = conversion.get_converter(datatypes.RangeInt0To100, datatypes.RangeFloat1)
        r4 = conv6(r3)
        self.assertIsInstance(r4, datatypes.RangeFloat1)
        self.assertAlmostEqual(r1, r4)

    def test_ranges_2(self) -> None:
        # And the other way round!

        v0 = 25
        conv0 = conversion.get_converter(int, datatypes.RangeInt0To100)
        r1 = conv0(v0)
        self.assertIsInstance(r1, datatypes.RangeInt0To100)
        self.assertEqual(datatypes.RangeInt0To100(25), r1)

        conv1 = conversion.get_converter(datatypes.RangeInt0To100, datatypes.RangeUInt8)
        r2 = conv1(r1)
        self.assertIsInstance(r2, datatypes.RangeUInt8)
        self.assertEqual(datatypes.RangeUInt8(64), r2)

        conv2 = conversion.get_converter(datatypes.RangeUInt8, datatypes.RangeFloat1)
        r3 = conv2(r2)
        self.assertIsInstance(r3, datatypes.RangeFloat1)
        self.assertAlmostEqual(0.25, float(r3), delta=0.01)

        conv3 = conversion.get_converter(datatypes.RangeFloat1, datatypes.RangeInt0To100)
        r4 = conv3(r3)
        self.assertEqual(r1, r4)

    def test_ranges_and_bool(self) -> None:
        r1 = datatypes.RangeUInt8(0)
        r2 = datatypes.RangeInt0To100(0)
        r3 = datatypes.RangeFloat1(0)
        r4 = datatypes.RangeUInt8(75)
        r5 = datatypes.RangeInt0To100(17)
        r6 = datatypes.RangeFloat1(0.6)
        self.assertFalse(conversion.get_converter(datatypes.RangeUInt8, bool)(r1))
        self.assertFalse(conversion.get_converter(datatypes.RangeInt0To100, bool)(r2))
        self.assertFalse(conversion.get_converter(datatypes.RangeFloat1, bool)(r3))
        self.assertTrue(conversion.get_converter(datatypes.RangeUInt8, bool)(r4))
        self.assertTrue(conversion.get_converter(datatypes.RangeInt0To100, bool)(r5))
        self.assertTrue(conversion.get_converter(datatypes.RangeFloat1, bool)(r6))

        self.assertEqual(datatypes.RangeUInt8(0), conversion.get_converter(bool, datatypes.RangeUInt8)(False))
        self.assertEqual(datatypes.RangeInt0To100(0), conversion.get_converter(bool, datatypes.RangeInt0To100)(False))
        self.assertEqual(datatypes.RangeFloat1(0), conversion.get_converter(bool, datatypes.RangeFloat1)(False))
        self.assertEqual(datatypes.RangeUInt8(255), conversion.get_converter(bool, datatypes.RangeUInt8)(True))
        self.assertEqual(datatypes.RangeInt0To100(100), conversion.get_converter(bool, datatypes.RangeInt0To100)(True))
        self.assertEqual(datatypes.RangeFloat1(1.0), conversion.get_converter(bool, datatypes.RangeFloat1)(True))

    def test_rgb(self) -> None:
        some_blue = datatypes.RGBFloat1(datatypes.RangeFloat1(.25),
                                        datatypes.RangeFloat1(.5),
                                        datatypes.RangeFloat1(.75))
        some_blue_int = conversion.get_converter(datatypes.RGBFloat1, datatypes.RGBUInt8)(some_blue)
        self.assertEqual(datatypes.RGBUInt8(datatypes.RangeUInt8(64),
                                            datatypes.RangeUInt8(128),
                                            datatypes.RangeUInt8(191)),
                         some_blue_int)

        some_blue_hsv = conversion.get_converter(datatypes.RGBFloat1, datatypes.HSVFloat1)(some_blue)
        self.assertAlmostEqual(210/360, some_blue_hsv.hue, delta=0.01)
        self.assertAlmostEqual(0.6667, some_blue_hsv.saturation, delta=0.01)
        self.assertAlmostEqual(0.75, some_blue_hsv.value, delta=0.01)

        some_blue_hsv2 = conversion.get_converter(datatypes.RGBUInt8, datatypes.HSVFloat1)(some_blue_int)
        self.assertAlmostEqual(210/360, some_blue_hsv2.hue, delta=0.01)
        self.assertAlmostEqual(0.6667, some_blue_hsv2.saturation, delta=0.01)
        self.assertAlmostEqual(0.75, some_blue_hsv2.value, delta=0.01)

        some_yellow_hsv = some_blue_hsv._replace(hue=datatypes.RangeFloat1(45/360))

        some_yellow = conversion.get_converter(datatypes.HSVFloat1, datatypes.RGBFloat1)(some_yellow_hsv)
        self.assertAlmostEqual(0.75, some_yellow.red, delta=0.01)
        self.assertAlmostEqual(0.625, some_yellow.green, delta=0.01)
        self.assertAlmostEqual(0.25, some_yellow.blue, delta=0.01)

        some_yellow_int = conversion.get_converter(datatypes.HSVFloat1, datatypes.RGBUInt8)(some_yellow_hsv)
        self.assertEqual(datatypes.RGBUInt8(datatypes.RangeUInt8(191),
                                            datatypes.RangeUInt8(159),
                                            datatypes.RangeUInt8(64)),
                         some_yellow_int)

        some_yellow2 = conversion.get_converter(datatypes.RGBUInt8, datatypes.RGBFloat1)(some_yellow_int)
        self.assertAlmostEqual(0.75, some_yellow2.red, delta=0.01)
        self.assertAlmostEqual(0.625, some_yellow2.green, delta=0.01)
        self.assertAlmostEqual(0.25, some_yellow2.blue, delta=0.01)

    def test_cct_and_rgbw(self) -> None:
        some_blue_int = datatypes.RGBUInt8(datatypes.RangeUInt8(64),
                                           datatypes.RangeUInt8(128),
                                           datatypes.RangeUInt8(191))
        some_blue_rgbw = conversion.get_converter(datatypes.RGBUInt8, datatypes.RGBWUInt8)(some_blue_int)
        self.assertEqual(datatypes.RGBWUInt8(some_blue_int, datatypes.RangeUInt8(0)),
                         some_blue_rgbw)

        some_blue_with_white = some_blue_rgbw._replace(white=datatypes.RangeUInt8(133))
        self.assertEqual(some_blue_int,
                         conversion.get_converter(datatypes.RGBWUInt8, datatypes.RGBUInt8)(some_blue_with_white))

        some_blue_with_cct = \
            conversion.get_converter(datatypes.RGBWUInt8, datatypes.RGBCCTUInt8)(some_blue_with_white)
        self.assertEqual(datatypes.RGBCCTUInt8(some_blue_int, datatypes.CCTUInt8(datatypes.RangeUInt8(133),
                                                                                 datatypes.RangeUInt8(133))),
                         some_blue_with_cct)

        some_blue_with_white_other_cct = some_blue_with_cct._replace(
            white=datatypes.CCTUInt8(datatypes.RangeUInt8(113), datatypes.RangeUInt8(153)))
        self.assertEqual(some_blue_with_white,
                         conversion.get_converter(datatypes.RGBCCTUInt8, datatypes.RGBWUInt8)
                         (some_blue_with_white_other_cct))

        self.assertEqual(datatypes.CCTUInt8(datatypes.RangeUInt8(113), datatypes.RangeUInt8(153)),
                         conversion.get_converter(datatypes.RGBCCTUInt8, datatypes.CCTUInt8)
                         (some_blue_with_white_other_cct))

        self.assertEqual(some_blue_int,
                         conversion.get_converter(datatypes.RGBCCTUInt8, datatypes.RGBUInt8)
                         (some_blue_with_cct))

        self.assertEqual(datatypes.RangeUInt8(133),
                         conversion.get_converter(datatypes.CCTUInt8, datatypes.RangeUInt8)
                         (datatypes.CCTUInt8(datatypes.RangeUInt8(113), datatypes.RangeUInt8(153))))
        self.assertEqual(datatypes.CCTUInt8(datatypes.RangeUInt8(133), datatypes.RangeUInt8(133)),
                         conversion.get_converter(datatypes.RangeUInt8, datatypes.CCTUInt8)
                         (datatypes.RangeUInt8(133)))


class ColorTests(unittest.TestCase):
    def test_rgb_scaling(self) -> None:
        some_blue = datatypes.RGBFloat1(datatypes.RangeFloat1(.25),
                                        datatypes.RangeFloat1(.5),
                                        datatypes.RangeFloat1(.75))

        some_dim_blue = some_blue.dimmed(datatypes.RangeFloat1(0.5))
        self.assertAlmostEqual(some_dim_blue.red, .125)
        self.assertAlmostEqual(some_dim_blue.green, .25)
        self.assertAlmostEqual(some_dim_blue.blue, .375)

        some_dim_blue2 = some_blue.dimmed(datatypes.RangeInt0To100(50))
        self.assertAlmostEqual(some_dim_blue2.red, .125, delta=0.01)
        self.assertAlmostEqual(some_dim_blue2.green, .25, delta=0.01)
        self.assertAlmostEqual(some_dim_blue2.blue, .375, delta=0.01)

        some_dim_blue3 = some_blue.dimmed(datatypes.RangeUInt8(127))
        self.assertAlmostEqual(some_dim_blue3.red, .125, delta=0.01)
        self.assertAlmostEqual(some_dim_blue3.green, .25, delta=0.01)
        self.assertAlmostEqual(some_dim_blue3.blue, .375, delta=0.01)

        some_blue_hsv = datatypes.HSVFloat1(datatypes.RangeFloat1(210/360),
                                            datatypes.RangeFloat1(.6667),
                                            datatypes.RangeFloat1(.75))

        some_dim_hsv_blue = some_blue_hsv.dimmed(datatypes.RangeFloat1(0.5))
        self.assertAlmostEqual(some_dim_hsv_blue.hue, 210/360)
        self.assertAlmostEqual(some_dim_hsv_blue.saturation, .6667)
        self.assertAlmostEqual(some_dim_hsv_blue.value, .375)

        some_dim_hsv_blue2 = some_blue_hsv.dimmed(datatypes.RangeInt0To100(50))
        self.assertAlmostEqual(some_dim_hsv_blue2.value, .375, delta=0.01)

        some_dim_hsv_blue3 = some_blue_hsv.dimmed(datatypes.RangeUInt8(127))
        self.assertAlmostEqual(some_dim_hsv_blue3.value, .375, delta=0.01)

    def test_rgbw_scaling(self) -> None:
        some_blue_with_white = datatypes.RGBWUInt8(datatypes.RGBUInt8(datatypes.RangeUInt8(64),
                                                                      datatypes.RangeUInt8(128),
                                                                      datatypes.RangeUInt8(191)),
                                                   datatypes.RangeUInt8(133))
        self.assertEqual(datatypes.RGBWUInt8(datatypes.RGBUInt8(datatypes.RangeUInt8(32),
                                                                datatypes.RangeUInt8(64),
                                                                datatypes.RangeUInt8(96)),
                                             datatypes.RangeUInt8(66)),
                         some_blue_with_white.dimmed(datatypes.RangeFloat1(0.5)))

        some_blue_with_cct = datatypes.RGBCCTUInt8(datatypes.RGBUInt8(datatypes.RangeUInt8(64),
                                                                      datatypes.RangeUInt8(128),
                                                                      datatypes.RangeUInt8(191)),
                                                   datatypes.CCTUInt8(datatypes.RangeUInt8(133),
                                                                      datatypes.RangeUInt8(118)))
        self.assertEqual(datatypes.RGBCCTUInt8(datatypes.RGBUInt8(datatypes.RangeUInt8(32),
                                                                  datatypes.RangeUInt8(64),
                                                                  datatypes.RangeUInt8(96)),
                                               datatypes.CCTUInt8(datatypes.RangeUInt8(66),
                                                                  datatypes.RangeUInt8(59))),
                         some_blue_with_cct.dimmed(datatypes.RangeFloat1(0.5)))



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""
This module contains all the predefined user interface widget (:class:`shc.web.WebPageItem`) classes, which can be
instantiated and added to :class:`WebPage` s in order to compose the web user interface.

In addition, there are descriptor classes, like the button classes (e.g. :class:`ToggleButton`) and the
:class:`ImageMapLabel`. Instances of these classes are used to create an interactive element within another widget.
They cannot be added to a ui page directly, but must be added to a compatible container element (like
:class:`ButtonGroup` or :class:`ImageMap`) instead. Still, they form the endpoint for the dynamic interaction and thus
are `Connectable` objects.

Please note that each widget or button instance must not be used multiple times (neither on the same nor on
different pages). To create two or more similar and synchronized widgets/buttons, create two instances with the same
settings and `connect` both of them to the same :class:`shc.Variable`.
"""


import abc
import enum
import itertools
import json
import pathlib
from os import PathLike
from typing import Any, Type, Union, Iterable, List, Generic, Tuple, TypeVar, Optional, Callable

import markupsafe
from markupsafe import Markup

from .interface import WebPageItem, WebDisplayDatapoint, WebActionDatapoint, jinja_env, WebConnectorContainer,\
    WebUIConnector, WebPage, WebServer
from ..base import T, ConnectableWrapper, Connectable
from ..conversion import SHCJsonEncoder
from ..datatypes import RangeFloat1, RGBUInt8


__all__ = [
    'icon',
    'Switch',
    'Select', 'EnumSelect',
    'ButtonGroup', 'ValueListButtonGroup', 'EnumButtonGroup',
    'ToggleButton', 'ValueButton', 'DisplayButton', 'StatelessButton',
    'TextDisplay', 'TextInput',
    'Slider',
    'HideRowBox', 'HideRow',
    'ColorChoser',
    'ImageMap', 'ImageMapLabel',
]


def icon(icon_name: str, label: str = '') -> Markup:
    """
    Create HTML markup for a Fontawesome/Semantic UI icon, to be used in PageItem labels, button labels, etc.

    :param icon_name: Name of the icon. See https://fomantic-ui.com/elements/icon.html for reference.
    :param label: Optional textual label to be placed along with the icon. The styling of the icon is slightly changed,
        when `label` is non-empty to optimize the spacing between icon an text.
    :return: (safe) HTML markup to be passed to a Jinja template (of a web page or web widget), e.g. via one of the
        `label` attributes.
    """
    return Markup('<i class="ui {}{} icon"></i>'.format("" if label else "fitted ", icon_name)) + label


class Switch(WebDisplayDatapoint[bool], WebActionDatapoint[bool], WebPageItem):
    """
    A `WebPageItem` showing a label and a right-aligned toggle switch.

    This widget is a `Connectable` object with type `bool`. It should be connected to a `Readable` object for
    initialization, e.g. a :class:`shc.Variable` of bool type.

    :param label: The label to be displayed near the switch. Either a plain string (which is automatically escaped for
        embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML, e.g.
        produced by :func:`icon`.
    :param color: Background color of the switch, when in 'on' position. The available colors are the same as for
        buttons. See https://fomantic-ui.com/elements/button.html#colored for reference.
    :param confirm_message: If provided, the user is asked for confirmation of each interaction with the switch with
        this message and "OK" and "Cancel" buttons. `confirm_values` may be used to restrict the confirmation to
        switching on or switching off.
    :param confirm_values: If `confirm_message` is given, this parameter specifies, which changes of the switch must be
        confirmed by the user. When set to `[True]`, only switching on (sending True) must be confirmed, when set to
        `[False]`, only switching off must be confirmed. Defaults to `[False, True]`, i.e. every change must be
        confirmed.
    """
    def __init__(self, label: Union[str, Markup], color: str = '', confirm_message: str = '',
                 confirm_values: Iterable[bool] = (False, True)):
        self.type = bool
        super().__init__()
        self.label = label
        self.color = color
        self.confirm_message = confirm_message
        self.confirm = confirm_values if confirm_message else ()

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/switch.htm').render_async(
            id=id(self), label=self.label, color=self.color,
            confirm_csv_int=",".join(str(int(v)) for v in self.confirm), confirm_message=self.confirm_message)


class Select(WebDisplayDatapoint[T], WebActionDatapoint[T], WebPageItem, Generic[T]):
    """
    A dropdown (HTML <select>) ui widget with label.

    This widget is a generic `Connectable` object. The type is determined by the values' type. It should be connected to
    a `Readable` object for initialization.

    :param options: List of options in the dropdown. Each option is a tuple of (value, label). All values must be of the
        same type. Each label can either be a plain string or a :class:`markupsafe.Markup` with pre-formatted HTML,
        e.g. from the :func:`icon` function. Since Semantic UI's JavaScript-based <select>-replacement is used, even
        complex HTML should be rendered properly.
    :param label: The label to be displayed near the dropdown. Either a plain string (which is automatically escaped for
        embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML, e.g.
        produced by :func:`icon`.
    """
    def __init__(self, options: List[Tuple[T, Union[str, Markup]]], label: str = ''):
        self.type = type(options[0][0])
        super().__init__()
        self.options = options
        self.label = label

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/select.htm').render_async(
            id=id(self), label=self.label, options=[(json.dumps(value, cls=SHCJsonEncoder), label)
                                                    for value, label in self.options])


class EnumSelect(Select):
    """
    Specialized version of :class:`Select` for choosing between the entries of a given enum type by their `name`.

    :param type_: The enum type. It is used as the `type` attribute for `connecting` and to generate the dropdown
        entries from its enum members.
    :param label: The label to be displayed near the dropdown. Either a plain string (which is automatically escaped for
        embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML, e.g.
        produced by :func:`icon`.
    """
    def __init__(self, type_: Type[enum.Enum], label: Union[str, Markup] = ''):
        values = [(entry, entry.name) for entry in type_]
        super().__init__(values, label)


Ti = TypeVar('Ti', int, float, str)


class TextInput(WebDisplayDatapoint[Ti], WebActionDatapoint[Ti], WebPageItem, Generic[Ti]):
    """
    An input field ui widget for numbers and strings with label.

    This widget is a generic `Connectable` object. The type is specified by the `type_` parameter. The `TextInput`
    should be connected to a `Readable` object for initialization.

    :param type_: The value type to be entered. Used as the `type` for connecting with other Connectables and to
        determine the HTML input type. Must be int, float or str.
    :param label: The label to be displayed near the input. Either a plain string (which is automatically escaped for
        embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML, e.g.
        produced by :func:`icon`.
    :param min: The minimal value to be entered (only useful if `type_` is int or float). Used as the HTML `min`
        attribute.
    :param max: The maximal value to be entered (only useful if `type_` is int or float). Used as the HTML `max`
        attribute.
    :param step: The default increment step when using the arrow buttons or arrow keys in the input field (only useful
        if `type_` is int or float). Used as the HTML `step` attribute.
    :param input_suffix: A plain string or HTML markup to be appended to the right of the input field, using Semantic
        UI's `labeled inputs <https://fomantic-ui.com/elements/input.html#labeled>`_.
    """
    def __init__(self, type_: Type[Ti], label: Union[str, Markup] = '', min: Optional[Ti] = None,
                 max: Optional[Ti] = None, step: Optional[Ti] = None, input_suffix: Union[str, Markup] = ''):
        self.type = type_
        super().__init__()
        self.label = label
        self.min: Optional[Ti] = min
        self.max: Optional[Ti] = max
        self.step: Optional[Ti] = step
        if self.step is None and issubclass(type_, int):
            self.step = 1
        self.input_type = "number" if issubclass(self.type, (int, float)) else "text"
        self.input_suffix = input_suffix

    def convert_from_ws_value(self, value: Any) -> Ti:
        return self.type(value)

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/textinput.htm').render_async(
            id=id(self), label=self.label, type=self.input_type, min=self.min, max=self.max, step=self.step,
            input_suffix=self.input_suffix)


class TextDisplay(WebDisplayDatapoint[T], WebPageItem):
    """
    A ui widget which simply displays the value of the connected `Connectable` object as text.

    This widget is a generic `Connectable` object. The type is specified by the `type_` parameter. The `TextInput`
    should be connected to a `Readable` object for initialization. In contrast to interactive widgets, it does only
    consume, but not produce values, so it's not `Subscribable`.

    The formatting of the value can be controlled via the `format_string` parameter, which is used with Python 3's
    `str.format()` method. To show the value with default formatting, use the '{}' format. Other format specifiers may
    be used, e.g. to specify the floating point accuracy::

        TextDisplay(float, '{:.2f}', "Current value of my_float").connect(my_float_variable)

    The format string may, of course, contain additional characters outside the format specifier to prepend and/or
    append static strings to the dynamically formatted value.
    See https://docs.python.org/3/library/string.html#formatstrings for a full reference of valid format strings.

    :param type_: The expected value type, used as the `type` attribute of this `Connectable`
    :param format: Either a Python string or safely escpaed HTML markup, which is used to format the value into a string
        representation using the `format()` method or otherwise a callable which transforms a value to a string or
        escaped HTML Markup (in form of a :class:`markupsafe.Markup` object).
    :param label: A label to be displayed left of the formatted value. Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    """
    def __init__(self, type_: Type[T], format: Union[str, Markup, Callable[[T], Union[str, Markup]]],
                 label: Union[str, Markup]):
        self.type = type_
        super().__init__()
        self.formatter: Callable[[T], Union[str, Markup]] = (
            (lambda x: format.format(x))  # type: ignore
            if isinstance(format, (str, Markup))
            else format)
        self.label = label

    def convert_to_ws_value(self, value: T) -> Any:
        return markupsafe.escape(self.formatter(value))

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/textdisplay.htm').render_async(id=id(self), label=self.label)


class Slider(WebDisplayDatapoint[RangeFloat1], WebActionDatapoint[RangeFloat1], WebPageItem):
    """
    A visual slider ui widget, labeled with a scale from 0%-100% to set range values.

    This widget is a generic `Connectable` object with type :class:`shc.datatypes.RangeFloat1`. It should be connected
    to a `Readable` object for initialization. Using the `convert` parameter of the `connect()` method, a `Slider`
    can be connected to `Connectable` objects of other default range types, like :class:`shc.datatypes.RangeUInt8`.

    :param label: The label to be displayed above the slider (left). Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    :param color: Background color of the slider and the upper right label showing the current value. Must be one of
        Semantic UI's predefined slider colors: https://fomantic-ui.com/modules/slider.html#colored
    """
    def __init__(self, label: Union[str, Markup] = '', color: str = ''):
        self.type = RangeFloat1
        super().__init__()
        self.label = label
        self.color = color

    def convert_from_ws_value(self, value: Any) -> RangeFloat1:
        return RangeFloat1(float(value))

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/slider.htm').render_async(
            id=id(self), label=self.label, color=self.color)


class ButtonGroup(WebPageItem):
    """
    A ui widget consisting of one or more (right-aligned) buttons with a label.

    The appearance (color, label, etc.) and possible interactions of each individual button is specified by a button
    descriptor (any subclass of :class:`AbstractButton`) for each button. These button descriptors form the SHC-side
    interface to the buttons to connect them with other `Connectable` objects. The `ButtonGroup` itself is not
    `Connectable`.

    :param label: The label to be shown left of the buttons
    :param buttons: List of button descriptors
    """
    def __init__(self, label: Union[str, Markup], buttons: Iterable["AbstractButton"]):
        super().__init__()
        self.label = label
        self.buttons = buttons

    def get_connectors(self) -> Iterable[WebUIConnector]:
        return self.buttons  # type: ignore

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/buttongroup.htm')\
            .render_async(label=self.label, buttons=self.buttons)


class AbstractButton(metaclass=abc.ABCMeta):
    """
    Abstract base class for button descriptor objects.

    Instances of Concrete subclasses can be passed to container widgets like :class:`ButtonGroup` and :class:`ImageMap`
    and specify the layout (color, label, etc.) and functionality of a button. They also form the SHC-side interface
    for interacting with the button, i.e. they are `Subscriable` and/or `Reading`+`Writable`.

    :var label: The label/text content of the button. Either a plain string (which should automatically be escaped for
        embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet and properly
        escaped HTML code.
    :var color: The color of the button. One of the Semantic UI button colors.
    :var stateful: If True, the button has an on/off state. It should only be shown fully colored, when in 'on' state.
        Additionally, it should be shown in 'loading' state until an initial value is received from the server.
    :var enabled: If False, the button should not be clickable (HTML disabled attribute)
    :var outline: If True (and `stateful==True`), the button should be shown with a colored outline in 'off' state
    :var confirm: A list of all values (True and/or False) which should be confirmed by the user after a click on the
        button, before being sent to the server. An emtpy list/tuple means no confirmation is required for any
        interaction with this button.
    :var confirm_message: The message to be shown in the confirm window, when a confirmation is required for an
        interaction with this button.
    """
    label: Union[str, Markup] = ''
    color: str = ''
    stateful: bool = True
    enabled: bool = True
    outline: bool = False
    confirm: Iterable[bool] = ()
    confirm_message: str = ""

    @property
    def confirm_csv_int(self) -> str:
        return ",".join(str(int(v)) for v in self.confirm)


jinja_env.tests['button'] = lambda item: isinstance(item, AbstractButton)


class StatelessButton(WebActionDatapoint[T], AbstractButton, Generic[T]):
    """
    Button descriptor for a stateless button, i.e. a button that has no visual on/off state and always sends the same
    value, when clicked.

    `StatelessButtons` are `Subscribable`, but not `Writable`. The `type` is the type of their `value`.

    :param value: The value to be published by this object when the button is clicked by the user
    :param label: The label/text content of the button. Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    :param color: The color of the button. One of the Semantic UI button colors. See
        https://fomantic-ui.com/elements/button.html#colored for reference. If not specified, the button is shown in the
        grey-ish default color. Since the button has no on/off state, the specified color is always shown.
    :param confirm_message: If not empty, the user must confirm each click on the button in a confirm windows with this
        message.
    :param outline: If True, the button is not shown fully colored, but only with its outline
        (`Semantic UI basic button <https://fomantic-ui.com/elements/button.html#basic>`_). Since the button has no
        on/off state, this holds all the time.
    """
    stateful = False

    def __init__(self, value: T, label: Union[str, Markup] = '', color: str = '', confirm_message: str = '',
                 outline: bool = False):
        self.type = type(value)
        super().__init__()
        self.value = value
        self.label = label
        self.color = color
        self.outline = outline
        if confirm_message:
            self.confirm = [False, True]
            self.confirm_message = confirm_message

    def convert_from_ws_value(self, value: Any) -> T:
        return self.value


class ValueButton(WebActionDatapoint[T], WebDisplayDatapoint[T], AbstractButton, Generic[T]):
    """
    Button descriptor for a button with a fixed value.

    This button publishes a fixed value upon every click, just like the :class:`StatelessButton`. Unlike a
    `StatelessButton`, it is dynamically lit up (shown in full color) when the current value of the connected object
    matches the fixed value of the button.

    The `type` of this `Connectable` object is determined from the `value`. The `ValueButton` should be connected
    to a `Readable` object for initialization of the ui.

    For creating multiple `ValueButtons` with different values for the same variable, take a look a the
    :class:`ValueListButtonGroup` widget.

    :param value: The value to be published by this object when the button is clicked by the user. Also the value which
        is compared to the `connected` object's current value to determine if the button should be lit up.
    :param label: The label/text content of the button. Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    :param color: The color of the button when it is lit up (i.e. `conncted` object's current value matches the
        `value`). Must be one of the Semantic UI button colors. See https://fomantic-ui.com/elements/button.html#colored
        for reference. Defaults to 'blue'.
    :param confirm_message: If not empty, the user must confirm each click on the button in a confirm windows with this
        message.
    :param outline: If True, the button is shown with a colored outline in its configured `color` **when not lit up**,
        instead of its grey-ish default appearance.
    """
    def __init__(self, value: T, label: Union[str, Markup] = '', color: str = 'blue',
                 confirm_message: str = '', outline: bool = False):
        self.type = type(value)
        super().__init__()
        self.value = value
        self.label = label
        self.color = color
        self.outline = outline
        if confirm_message:
            self.confirm = [False, True]
            self.confirm_message = confirm_message

    def convert_from_ws_value(self, value: Any) -> T:
        return self.value

    def convert_to_ws_value(self, value: T) -> Any:
        return value == self.value


class ToggleButton(WebActionDatapoint[bool], AbstractButton, WebDisplayDatapoint[bool]):
    """
    Button descriptor for an on/off button.

    Click such a button once to turn it 'on', click it again to turn it 'off'. The button is lit up (fully colored),
    when in 'on' state (the `connected` object's value is `True`) and shown in the grey-ish default color, when in 'off'
    state (unless `outline==True`). When clicked in 'on' state, a `False` value is published and vice versa.

    A `ToggleButton` is a `Connectable` object with type `bool´. It should be connected to a `Readable` object for
    initialization of the ui – otherwise it will show a spinner animation until the first True/False value is received.

    :param label: The label/text content of the button. Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    :param color: The color of the button when in 'on' state. Must be one of the Semantic UI button colors.
        See https://fomantic-ui.com/elements/button.html#colored for reference. Defaults to 'blue'.
    :param confirm_message: If provided, the user is asked for confirmation of each interaction with the button with
        this message and "OK" and "Cancel" buttons. `confirm_values` may be used to restrict the confirmation to
        switching on or switching off.
    :param confirm_values: If `confirm_message` is given, this parameter specifies, which changes of the button must be
        confirmed by the user. When set to `[True]`, only switching on (sending True, i.e. clicking in 'off' state) must
        be confirmed; when set to `[False]`, only switching off must be confirmed. Defaults to `[False, True]`, i.e.
        every click must be confirmed.
    :param outline: If True, the button is shown with a colored outline in its configured `color` **in 'off' state**,
        instead of its grey-ish default appearance.
    """
    def __init__(self, label: Union[str, Markup] = '', color: str = 'blue', confirm_message: str = '',
                 confirm_values: Iterable[bool] = (False, True), outline: bool = False):
        self.type = bool
        super().__init__()
        self.label = label
        self.color = color
        self.outline = outline
        if confirm_message:
            self.confirm_message = confirm_message
            self.confirm = confirm_values


class DisplayButton(WebDisplayDatapoint[T], AbstractButton, Generic[T]):
    """
    Button descriptor for a read-only (non-clickable) button.

    A `DisplayButton` behaves similar to a :class:`ValueButton`: It has a preconfigured fixed `value` and is lit up
    (displayed in full color) when the `connected` object's value equals this fixed value. However the `DisplayButton`
    is not clickable. It is a pure display widget, which does not provide user interaction. This, on the SHC-side it
    is not `Subscribable`.

    The `type` of this `Connectable` object is determined from the `value`. The `DisplayButton` should be connected
    to a `Readable` object for initialization of the ui.

    :param value: The value which is compared to the `connected` object's current value to determine if the button
        should be lit up. Defaults to `True`, resulting in a behaviour of a disabled :class:`ToggleButton`.
    :param label: The label/text content of the button. Either a plain string (which is automatically
        escaped for embedding in HTML) or a :class:`markupsafe.Markup` object, which may contain pre-formattet HTML,
        e.g. produced by :func:`icon`.
    :param color: The color of the button when in 'on' state. Must be one of the Semantic UI button colors.
        See https://fomantic-ui.com/elements/button.html#colored for reference. Defaults to 'blue'.
    :param outline: If True, the button is shown with a colored outline in its configured `color` **when not lit up**,
        instead of its grey-ish default appearance.
    """
    enabled = False

    def __init__(self, value: T = True,  label: Union[str, Markup] = '',  # type: ignore
                 color: str = 'blue', outline: bool = False):
        self.type = type(value)
        super().__init__()
        self.value = value
        self.label = label
        self.color = color
        self.outline = outline

    def convert_to_ws_value(self, value: T) -> bool:
        return value == self.value


class ValueListButtonGroup(ButtonGroup, ConnectableWrapper[T], Generic[T]):
    """
    A derived :class:`ButtonGroup` widget to simplify creating a list of :class:`ValueButtons` for the same variable.

    For convenience, this object is a :class:`ConnectableWrapper`. This means, you can use the :meth:`connect` method
    with another `Connectable` object (like a `Variable`), to connect all the contained `ValueButtons` to that object
    at once.

    :param values: A list of `(value, label)` tuples. For each tuple, a :class:`ValueButton` is created. All values must
        be of the same type.
    :param label: The label of the :class:`ButtonGroup`
    :param color: A common `color` for all buttons
    :param confirm_message: A common `confirm_message` for all buttons
    """
    def __init__(self, values: List[Tuple[T, Union[str, Markup]]], label: Union[str, Markup],
                 color: str = 'blue', confirm_message: str = ''):
        buttons = [ValueButton(value=v[0], label=v[1], color=color, confirm_message=confirm_message) for v in values]
        super().__init__(label, buttons)

    def connect(self, *args, **kwargs):
        for button in self.buttons:
            assert isinstance(button, Connectable)
            button.connect(*args, **kwargs)
        return self


class EnumButtonGroup(ValueListButtonGroup):
    """
    A derived of :class:`ValueListButtonGroup`, which is a :class:`ButtonGroup` of :class:`ValueButtons`

    This specialized variant takes an enum type (derived from :class:`enum.Enum`) and creates a :class:`ValueButton` for
    each entry/member of that enum, with the members' names as button labels. Like `ValueListButtonGroup`, this object
    is a :class:`ConnectableWrapper`, so you can use the :meth:`connect` to connect all the contained `ValueButtons` to
    another `Connectable` object at once.

    :param type_: The enum type to create the value buttons for
    :param label: The label of the :class:`ButtonGroup`
    :param color: A common `color` for all buttons
    :param confirm_message: A common `confirm_message` for all buttons
    """
    def __init__(self, type_: Type[enum.Enum], label: Union[str, Markup], color: str = 'blue',
                 confirm_message: str = ''):
        values = [(entry, entry.name) for entry in type_]
        super().__init__(values, label, color, confirm_message)


class HideRowBox(WebPageItem):
    def __init__(self, rows: List["HideRow"]):
        self.rows = rows

    def get_connectors(self) -> Iterable[WebUIConnector]:
        return itertools.chain.from_iterable(row.get_connectors() for row in self.rows)

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/hiderowbox.htm').render_async(rows=self.rows)


class HideRow(WebDisplayDatapoint[bool], WebConnectorContainer):
    def __init__(self, label: Union[str, Markup], button: Optional[AbstractButton] = None,
                 color: str = 'blue',):
        self.type = bool
        super().__init__()
        self.label = label
        self.button = button
        self.color = color

    def get_connectors(self) -> Iterable[WebUIConnector]:
        if self.button:
            yield self.button  # type: ignore
        yield self


class ColorChoser(WebActionDatapoint[RGBUInt8], WebDisplayDatapoint[RGBUInt8], WebPageItem):
    type = RGBUInt8

    async def render(self) -> str:
        # TODO add label
        return await jinja_env.get_template('widgets/colorchoser.htm').render_async(id=id(self))


ImageMapItem = Union[AbstractButton, "ImageMapLabel"]


class ImageMap(WebPageItem):
    def __init__(self, image: Union[PathLike, str],
                 items: Iterable[Union[Tuple[float, float, ImageMapItem],
                                       Tuple[float, float, ImageMapItem, List[WebPageItem]]]],
                 max_width: Optional[int] = None):
        super().__init__()
        self.image = image
        self.image_url: str = ''
        # Allow using external images: If `image` is an absolute URI, simply use it as the img src, instead of trying
        # to serve it via our http server
        if isinstance(image, str) and '://' in image:
            self.image_url = image

        self.max_width = max_width
        self.items: List[Tuple[float, float, ImageMapItem, List[WebPageItem]]]\
            = [item if len(item) >= 4 else (item[0], item[1], item[2], [],)  # type: ignore # (MyPy does not get it ...)
               for item in items]

    def register_with_server(self, _page: WebPage, server: WebServer) -> None:
        if not self.image_url:
            self.image_url = server.serve_static_file(pathlib.Path(self.image))

    def get_connectors(self) -> Iterable[WebUIConnector]:
        for x, y, item, sub_items in self.items:
            if isinstance(item, WebUIConnector):
                yield item
            yield from itertools.chain.from_iterable(i.get_connectors() for i in sub_items)

    async def render(self) -> str:
        return await jinja_env.get_template('widgets/imagemap.htm')\
            .render_async(items=self.items, image_url=self.image_url, max_width=self.max_width)


class ImageMapLabel(WebDisplayDatapoint[T]):
    def __init__(self, type_: Type[T], format_string: str = "{}", color: str = ""):
        self.type = type_
        super().__init__()
        self.color = color
        self.format_string = format_string

    def convert_to_ws_value(self, value: T) -> Any:
        return self.format_string.format(value)


jinja_env.tests['imageMapLabel'] = lambda item: isinstance(item, ImageMapLabel)



# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import logging
from typing import Set, Optional

from ..supervisor import AbstractInterface, interface_failure, InterfaceStatus, ServiceStatus

logger = logging.getLogger(__name__)


class SupervisedClientInterface(AbstractInterface, metaclass=abc.ABCMeta):
    """
    Abstract base class for client interfaces, providing run task supervision and automatic reconnects

    This class can be used as a base class for client interface implementations to simplify error handling and automatic
    reconnection attempts. The core of its functionality is the `_supervise` Task. It is started at interface `start()`
    and runs until shutdown of the interface. It starts and supervises the implementation-specific :meth:`_run` Task,
    which typically contains a loop for handling incoming messages. In addition, it calls :meth:`_connect` **before**
    and :meth:`_subscribe` **after** starting the run task. If the run task exits unexpectedly and
    :attr:`auto_reconnect` is enabled, reconnecting the interface via `_connect`, `_run` and `_subscribe` is attempted.
    For shutting down the interface (and stopping the run task in case of a subscribe error), `_disconnect` must be
    implemented in such a way, that it shuts down the run task.
    """
    def __init__(self, auto_reconnect: bool = True, failsafe_start: bool = False):
        super().__init__()
        """
        :param auto_reconnect: If True (default), the supervisor tries to reconnect the interface automatically with
            exponential backoff (`backoff_base` * `backoff_exponent` ^ n seconds sleep), when `_run` exits unexpectedly
            or any of _connect, _run or _subscribe raise an exception. Otherwise, the complete SHC system is shut down
            on connection errors.
        :param failsafe_start: If True and auto_reconnect is True, the interface allows SHC to start up, even if the
            `_connect` or `_subscribe` fails in the first try. The connection is retried in background with exponential
            backoff (see `auto_reconnect` option). Otherwise (default), the first connection attempt on startup is not
            retried and will raise an exception from `start()` on failure, even if `auto_reconnect` is True.
        """
        self.auto_reconnect = auto_reconnect
        self.failsafe_start = failsafe_start and auto_reconnect
        self.backoff_base = 1.0  #: First wait interval for exponential backoff in seconds
        self.backoff_exponent = 1.25  #: Multiplier for wait intervals for exponential backoff
        self._supervise_task: Optional[asyncio.Task] = None
        loop = asyncio.get_event_loop()
        self._started = loop.create_future()
        self._stopping = asyncio.Event()
        self._running = asyncio.Event()
        self._last_error: str = "Interface has not been started yet"

    async def start(self) -> None:
        logger.debug("Starting supervisor task for interface %s and waiting for it to come up ...", self)
        self._supervise_task = asyncio.create_task(self._supervise())
        await self._started

    async def stop(self) -> None:
        if self._stopping.is_set():
            return
        self._stopping.set()
        await self._disconnect()
        if self._supervise_task is not None:
            await self._supervise_task

    async def get_status(self) -> InterfaceStatus:
        return InterfaceStatus(ServiceStatus.OK if self._running.is_set() else ServiceStatus.CRITICAL,
                               self._last_error if not self._running.is_set() else "",
                               {})

    async def wait_running(self, timeout: Optional[float] = None) -> None:
        """
        Wait for the interface to be running.

        Attention: This must be called *after* :meth:`start` has initially been called (not neccessarily after it has
        returned).

        :param timeout: If given, this method will raise an :class:`asyncio.TimeoutError` after the given timeout in
            seconds, if the interface has not come up by this time.
        """
        await asyncio.wait_for(self._running.wait(), timeout)

    @abc.abstractmethod
    async def _run(self) -> None:
        """
        Entrypoint for the run task, which handles messages etc. while the connection is active.

        This coroutine is started in a separate task *after* :meth:`_connect` has completed sucessfully. As soon as it
        is ready, it must set the :attr:`_running` event. Only then the :meth:`_subscribe` method is called and the
        startup of the interface is reported as finished.

        The `_run` coroutine should be stoppable by calling :meth:`_disconnect` (i.e. it should return or raise an
        exception when `_disconnect` is called). If in doubt, just add a new :class:`asyncio.Event` and use
        :func:`asyncio.wait` with the event's `.wait()` method and your original future in all places, where you need to
        await a future. In addition, `_run` should return or raise an exception when a client error occurs, in order to
        trigger a reconnect attempt.
        """
        pass

    @abc.abstractmethod
    async def _connect(self) -> None:
        """
        This coroutine is run to connect the client.

        This will happen at start up and after any error (if auto_reconnect is enabled). In case of an error, *no*
        disconnect is attempted before calling `_connect`. Thus, this coroutine should be able to handle any connection
        state (not yet connected, not connected due to failed `_connect` attempt, broken connection, open connection
        with failed `_subscribe` call).

        This method is called *before* starting the :meth:`_run` task.
        """
        pass

    @abc.abstractmethod
    async def _disconnect(self) -> None:
        """
        This coroutine is called to disconnect the client and stop the _run task.

        This may happen either to shut down the interface (when stop() is called by the supervisor) or when an error
        occurs during :meth:`_subscribe` or when any error occurred and no auto_reconnect is attempted. Thus, disconnect
        should be able to shut down the client in a failed state as well. It should also be idempotent, i.e. allow to be
        called multiple times without reconnect. This method should not raise Exceptions but instead try its best to
        shut down the interface.

        Calling this coroutine must somehow stop the run task. I.e. :meth:`_run` should return or raise an exception
        shortly afterwards.
        """
        pass

    @abc.abstractmethod
    async def _subscribe(self) -> None:
        """
        This coroutine is called *after* :meth:`connecting <_connect>` the client and starting the :meth:`_run` task.

        It can be used to subscribe to topics, send initialization messages, etc. It will be called again after a
        reconnect, when an error occurs and `auto_reconnect` is enabled.
        """
        pass

    async def _supervise(self) -> None:
        sleep_interval = self.backoff_base

        while True:
            exception = None
            wait_stopping = asyncio.create_task(self._stopping.wait())
            try:
                # Connect
                logger.debug("Running _connect for interface %s ...", self)
                connect_task = asyncio.create_task(self._connect())
                # TODO timeout
                done, _ = await asyncio.wait((connect_task, wait_stopping), return_when=asyncio.FIRST_COMPLETED)
                if connect_task not in done:
                    logger.debug("Interface %s stopped before _connect finished", self)
                    connect_task.cancel()
                connect_task.result()  # raise exception if any

                # Start _run task and wait for _running
                logger.debug("Starting _run task for interface %s ...", self)
                run_task = asyncio.create_task(self._run())
                wait_running = asyncio.create_task(self._running.wait())
                # TODO timeout
                logger.debug("Waiting for interface %s to report it is running ...", self)
                done, _ = await asyncio.wait((wait_running, run_task), return_when=asyncio.FIRST_COMPLETED)
                if wait_running not in done:
                    wait_running.cancel()
                    if run_task not in done:  # This should not happen (without timeout)
                        await self._disconnect()
                        await run_task
                    raise RuntimeError("Run task stopped before _running has been set")

                # Subscribe
                logger.debug("Starting _subscribe task for interface %s ...", self)
                subscribe_task = asyncio.create_task(self._subscribe())
                # TODO timeout
                done, _ = await asyncio.wait((subscribe_task, run_task), return_when=asyncio.FIRST_COMPLETED)
                if subscribe_task not in done:
                    if run_task not in done:  # This should not happen (without timeout)
                        await self._disconnect()
                        await run_task
                    raise RuntimeError("Run task stopped before _subscribe task finished")
                subscribe_exception = subscribe_task.exception()
                if subscribe_exception is not None:
                    await self._disconnect()
                    try:
                        await run_task
                    except Exception as e:
                        logger.debug("Ignoring Exception %s in run task of interface %s, during shutdown due to "
                                     "exception in _subscribe task", repr(e), self)
                    raise subscribe_exception

                logger.debug("Starting up interface %s completed", self)
                if not self._started.done():
                    self._started.set_result(None)
                # Reset reconnect backoff interval
                sleep_interval = self.backoff_base

                # Wait for run task to return (due to stopping or error)
                await run_task

            except Exception as e:
                exception = e
                pass
            finally:
                wait_stopping.cancel()
            self._running.clear()

            # If we have not been started successfully yet, report startup as finished (if failsafe) or report startup
            # error and quit
            if not self._started.done():
                if self.failsafe_start:
                    self._started.set_result(None)
                else:
                    logger.debug("Startup of interface %s has not been finished due to exception", self)
                    self._started.set_exception(exception if exception is not None else asyncio.CancelledError())
                    await self._disconnect()
                    return

            # Return if we are stopping
            if self._stopping.is_set():
                if exception:
                    logger.debug("Ignoring exception %s in interface %s while stopping", repr(exception), self)
                return

            # Shut down SHC if no auto_reconnect shall be attempted
            if not self.auto_reconnect:
                if exception:
                    logger.critical("Error in interface %s:", exc_info=exception)
                else:
                    logger.critical("Unexpected shutdown of interface %s", self)
                asyncio.create_task(interface_failure(repr(self)))
                return

            if exception:
                logger.error("Error in interface %s. Attempting reconnect ...", self, exc_info=exception)
                self._last_error = str(exception)
            else:
                logger.error("Unexpected shutdown of interface %s. Attempting reconnect ...", self)
                self._last_error = "Unexpected shutdown of interface"

            # Sleep before reconnect
            logger.info("Waiting %s seconds before reconnect of interface %s ...", sleep_interval, self)
            wait_stopping = asyncio.create_task(self._stopping.wait())
            done, _ = await asyncio.wait((wait_stopping,), timeout=sleep_interval)
            if wait_stopping in done:
                logger.debug("Stopped interface %s while waiting for reconnect", self)
                return
            else:
                wait_stopping.cancel()
            sleep_interval *= self.backoff_exponent
            logger.info("Attempting reconnect of interface %s ...", self)



import unittest


class InterfaceImportTest(unittest.TestCase):
    def test_knx(self):
        import shc.interfaces.knx

    def test_dmx(self):
        import shc.interfaces.dmx

    def test_midi(self):
        import shc.interfaces.midi



# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
""" Outsourced C function bindings for volume conversion in the pulse interface datatypes.

Loading this module requires libpulse to be installed.
"""
import ctypes as c
import ctypes.util
from pulsectl._pulsectl import PA_CVOLUME, PA_CHANNEL_MAP


libpulse = c.CDLL(ctypes.util.find_library('libpulse') or 'libpulse.so.0')

pa_volume_t = c.c_uint32
pa_cvolume_max = libpulse.pa_cvolume_max
pa_cvolume_max.argtypes = [c.POINTER(PA_CVOLUME)]
pa_cvolume_max.restype = pa_volume_t
pa_cvolume_get_balance = libpulse.pa_cvolume_get_balance
pa_cvolume_get_balance.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP)]
pa_cvolume_get_balance.restype = c.c_float
pa_cvolume_get_fade = libpulse.pa_cvolume_get_fade
pa_cvolume_get_fade.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP)]
pa_cvolume_get_fade.restype = c.c_float
pa_cvolume_get_lfe_balance = libpulse.pa_cvolume_get_lfe_balance
pa_cvolume_get_lfe_balance.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP)]
pa_cvolume_get_lfe_balance.restype = c.c_float
pa_cvolume_set_balance = libpulse.pa_cvolume_set_balance
pa_cvolume_set_balance.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP), c.c_float]
pa_cvolume_set_balance.restype = None
pa_cvolume_set_fade = libpulse.pa_cvolume_set_fade
pa_cvolume_set_fade.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP), c.c_float]
pa_cvolume_set_fade.restype = None
pa_cvolume_set_lfe_balance = libpulse.pa_cvolume_set_lfe_balance
pa_cvolume_set_lfe_balance.argtypes = [c.POINTER(PA_CVOLUME), c.POINTER(PA_CHANNEL_MAP), c.c_float]
pa_cvolume_set_lfe_balance.restype = None
pa_cvolume_scale = libpulse.pa_cvolume_scale
pa_cvolume_scale.argtypes = [c.POINTER(PA_CVOLUME), pa_volume_t]
pa_cvolume_scale.restype = None



# Copyright 2020-2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import asyncio
import itertools
import json
import logging
import os
import pathlib
import weakref
from json import JSONDecodeError
from typing import Dict, Iterable, Union, List, Set, Any, Optional, Tuple, Generic, Type, Callable

import aiohttp.web
import jinja2
from aiohttp import WSCloseCode

from ..base import Reading, T, Writable, Subscribable
from ..conversion import SHCJsonEncoder, from_json
from ..supervisor import get_interfaces, AbstractInterface, ServiceStatus

logger = logging.getLogger(__name__)

jinja_env = jinja2.Environment(
    loader=jinja2.PackageLoader('shc.web', 'templates'),
    autoescape=jinja2.select_autoescape(['html', 'xml']),
    enable_async=True,
    trim_blocks=True,
    lstrip_blocks=True,
)
jinja_env.filters['id'] = id

LastWillT = Tuple["WebApiObject[T]", T]


class WebServer(AbstractInterface):
    """
    A SHC interface to provide the web user interface and a REST+websocket API for interacting with Connectable objects.

    :param host: The listening host. Use "" to listen on all interfaces or "localhost" to listen only on the
        loopback interface.
    :param port: The port to listen on
    :param index_name: Name of the `WebPage`, the root URL redirects to. If None, the root URL returns an HTTP 404.
    :param root_url: The base URL, at witch the user will reach this server. Used to construct internal links. May
        be an absolute URI (like "https://myhost:8080/shc") or an absolute-path reference (like "/shc"). Defaults
        to "". Note: This does not affect the routes of this HTTP server. It is only relevant, if you use an HTTP
        reverse proxy in front of this application, which serves the application in a sub path.
    :param title_formatter: A format string or format function to create the full HTML title, typically shown as browser
        tab title, from a web page's title. If it is a string, it should have one positional format placeholder
        (``{}``)
    :param enable_monitoring: If True (default), the monitoring endpoint at `/monitoring` is enabled to allow monitoring
        the interfaces' status in the UI or from a monitoring system
    """
    def __init__(self, host: str, port: int, index_name: Optional[str] = None, root_url: str = "",
                 title_formatter: Union[str, Callable[[str], str]] = "{} | SHC", enable_monitoring: bool = True):
        super().__init__()
        self.host = host
        self.port = port
        self.index_name = index_name
        self.root_url = root_url
        self.title_formatter = (title_formatter
                                if callable(title_formatter)
                                else lambda x: title_formatter.format(x))  # type: ignore

        # a dict of all `WebPage`s by their `name` for rendering them in the `_page_handler`
        self._pages: Dict[str, WebPage] = {}
        # a dict of all `WebConnector`s by their Python object id for routing incoming websocket mesages
        self.connectors: Dict[int, WebUIConnector] = {}
        # a dict of all `WebApiObject`s by their name for handling incoming HTTP requests and subscribe messages
        self._api_objects: Dict[str, WebApiObject] = {}
        # a set of all open websockets to close on graceful shutdown
        self._websockets: weakref.WeakSet[aiohttp.web.WebSocketResponse] = weakref.WeakSet()
        # a set of all open tasks to close on graceful shutdown
        self._associated_tasks: weakref.WeakSet[asyncio.Task] = weakref.WeakSet()
        # last will (object, value) per API websocket client (if set)
        self._api_ws_last_will: Dict[aiohttp.web.WebSocketResponse, LastWillT] = {}
        # data structure of the user interface's main menu
        # The structure looks as follows:
        # [('Label', 'icon', 'page_name'),
        #  ('Submenu label', None, [
        #     ('Label 2', 'icon', 'page_name2'), ...
        #   ]),
        #  ...]
        self.ui_menu_entries: List[Tuple[str, Optional[str], Union[str, List[Tuple[str, Optional[str], str]]]]] = []
        # List of all static js URLs to be included in the user interface pages
        self._js_files = [
            "/static/pack/main.js",
        ]
        # List of all static css URLs to be included in the user interface pages
        self._css_files = [
            "/static/pack/main.css",
        ]
        # A dict of all static files served by the application. Used to make sure, any of those is only served at one
        # path, when added via `serve_static_file()` multiple times.
        self.static_files: Dict[pathlib.Path, str] = {}

        # The actual aiohttp web app
        self._app = aiohttp.web.Application()
        self._app.add_routes([
            aiohttp.web.get("/", self._index_handler),
            aiohttp.web.get("/page/{name}/", self._page_handler, name='show_page'),
            aiohttp.web.get("/ws", self._ui_websocket_handler),
            aiohttp.web.static('/static', os.path.join(os.path.dirname(__file__), 'static')),
            aiohttp.web.get("/api/v1/ws", self._api_websocket_handler),
            aiohttp.web.get("/api/v1/object/{name}", self._api_get_handler),
            aiohttp.web.post("/api/v1/object/{name}", self._api_post_handler),
        ])
        if enable_monitoring:
            self._app.add_routes([aiohttp.web.get("/monitoring", self._monitoring_handler)])

    async def start(self) -> None:
        logger.info("Starting up web server on %s:%s ...", self.host, self.port)
        for connector in itertools.chain.from_iterable(page.get_connectors() for page in self._pages.values()):
            self.connectors[id(connector)] = connector
        for api_object in self._api_objects.values():
            api_object.start()
        self._runner = aiohttp.web.AppRunner(self._app)
        await self._runner.setup()
        site = aiohttp.web.TCPSite(self._runner, self.host, self.port)
        await site.start()

    async def stop(self) -> None:
        logger.info("Closing open websockets ...")
        for ws in set(self._websockets):
            await ws.close(code=WSCloseCode.GOING_AWAY, message=b'Server shutdown')
        for task in set(self._associated_tasks):
            task.cancel()
        logger.info("Cleaning up AppRunner ...")
        await self._runner.cleanup()

    def page(self, name: str, title: Optional[str] = None, menu_entry: Union[bool, str] = False,
             menu_icon: Optional[str] = None, menu_sub_label: Optional[str] = None, menu_sub_icon: Optional[str] = None
             ) -> "WebPage":
        """
        Create a new WebPage with a given name.

        If there is already a page with that name existing, it will be returned.

        :param name: The `name` of the page, which is used in the page's URL to identify it.
        :param title: The title/heading of the page. If not given, the name is used.
        :param menu_entry: If True (or a none-empty string) and this is a new page, an entry in the main menu will be
            created for the page. If `menu_entry` is a string, it will be used as the label, otherwise, the title will
            be used as a label.
        :param menu_icon: If given, the menu entry is prepended with the named icon
        :param menu_sub_label: If given, the menu entry is labeled with `menu_sub_label` and added to a submenu, labeled
            with `menu_entry` (and `menu_icon`, if given).
        :param menu_sub_icon: If given and `menu_sub_label` is given, the named icon is prepended to the submenu entry.
        :return: The new WebPage object or the existing WebPage object with that name
        :raises ValueError: If `menu_entry` is not False and there is already a menu entry with the same label (or a
            submenu entry with the same two labels)
        """
        if name in self._pages:
            return self._pages[name]
        else:
            if not title:
                title = name
            page = WebPage(self, name, title)
            self._pages[name] = page
            if menu_entry:
                self.add_menu_entry(name,
                                    menu_entry if isinstance(menu_entry, str) else title,
                                    menu_icon, menu_sub_label, menu_sub_icon)
            return page

    def add_menu_entry(self, page_name: str, label: str, icon: Optional[str] = None, sub_label: Optional[str] = None,
                       sub_icon: Optional[str] = None) -> None:
        """
        Create an entry for a named web UI page in the web UI's main navigation menu.

        The existence of the page is not checked, so menu entries can be created before the page has been created.

        :param page_name: The name of the page (link target)
        :param label: The label of the entry (or the submenu to place the entry in) in the main menu
        :param icon: If given, the menu entry is prepended with the named icon
        :param sub_label: If given, the menu entry is labeled with `sub_label` and added to a submenu, labeled with
            `label` (and `icon`, if given).
        :param sub_icon: If given and `menu_sub_label` is given, the named icon is prepended to the submenu entry.
        :raises ValueError: If there is already a menu entry with the same label (or a submenu entry with the same two
            labels)
        """
        existing_entry = next((e for e in self.ui_menu_entries if e[0] == label), None)
        if not sub_label:
            if existing_entry:
                raise ValueError("UI main menu entry with label {} exists already. Contents: {}"
                                 .format(label, existing_entry[2]))
            self.ui_menu_entries.append((label, icon, page_name))

        elif existing_entry:
            if not isinstance(existing_entry[2], list):
                raise ValueError("Existing UI main menu entry with label {} is not a submenu but a link to page {}"
                                 .format(label, existing_entry[2]))
            existing_entry[2].append((sub_label, sub_icon, page_name))

        else:
            self.ui_menu_entries.append((label, icon, [(sub_label, sub_icon, page_name)]))

    def api(self, type_: Type, name: str) -> "WebApiObject":
        """
        Create a new API endpoint with a given name and type.

        :param type_: The value type of the API endpoint object. Used as the *Connectable* object's `type` attribute and
            for JSON-decoding/encoding the values transmitted via the API.
        :param name: The name of the API object, which is the distinguishing part of the REST-API endpoint URL and used
            to identify the object in the websocket API.
        :return: A *Connectable* object that represents the API endpoint.
        """
        if name in self._api_objects:
            existing = self._api_objects[name]
            if existing.type is not type_:
                raise TypeError("Type {} does not match type {} of existing API object with same name"
                                .format(type_, existing.type))
            return existing
        else:
            api_object = WebApiObject(type_, name)
            self._api_objects[name] = api_object
            return api_object

    async def _index_handler(self, _request: aiohttp.web.Request) -> aiohttp.web.Response:
        if not self.index_name:
            raise aiohttp.web.HTTPNotFound()
        raise aiohttp.web.HTTPFound(self.root_url + str(self._app.router['show_page'].url_for(name=self.index_name)))

    async def _page_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        try:
            page = self._pages[request.match_info['name']]
        except KeyError:
            raise aiohttp.web.HTTPNotFound()

        html_title = self.title_formatter(page.title)
        template = jinja_env.get_template('page.htm')
        body = await template.render_async(title=page.title, segments=page.segments, menu=self.ui_menu_entries,
                                           root_url=self.root_url, js_files=self._js_files, css_files=self._css_files,
                                           server_token=id(self), html_title=html_title)
        return aiohttp.web.Response(body=body, content_type="text/html", charset='utf-8')

    async def _ui_websocket_handler(self, request: aiohttp.web.Request) -> aiohttp.web.WebSocketResponse:
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        self._websockets.add(ws)

        msg: aiohttp.WSMessage
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await self._ui_websocket_dispatch(ws, msg)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    logger.info('UI websocket connection closed with exception %s', ws.exception())
        finally:
            logger.debug('UI websocket connection closed')
            # Make sure the websocket is removed as a subscriber from all WebDisplayDatapoints
            self._websockets.discard(ws)
            for connector in self.connectors.values():
                connector.websocket_close(ws)
            return ws

    async def _ui_websocket_dispatch(self, ws: aiohttp.web.WebSocketResponse, msg: aiohttp.WSMessage) -> None:
        message = msg.json()
        if 'serverToken' in message:
            # Detect server restarts (if serverToken of client's page is different from our current server id) and
            # ask client to reload page.
            if message['serverToken'] != id(self):
                logger.debug("Client's serverToken %s does not match our id. Asking for reload.",
                             message['serverToken'])
                asyncio.create_task(ws.send_json({'reload': True}))
            return

        try:
            connector = self.connectors[message["id"]]
        except KeyError:
            logger.error("Could not route message from websocket to connector, since no connector with id %s is "
                         "known.", message['id'])
            return
        if 'v' in message:
            # We don't need to do this in an asynchronous task, since the from_websocket method uses asynchronous
            # publishing
            connector.from_websocket(message['v'], ws)
        elif 'sub' in message:
            asyncio.create_task(connector.websocket_subscribe(ws))
        else:
            logger.warning("Don't know how to handle websocket message: %s", message)

    async def _api_websocket_handler(self, request: aiohttp.web.Request) -> aiohttp.web.WebSocketResponse:
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)
        self._websockets.add(ws)

        msg: aiohttp.WSMessage
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    # It does not make sense to avoid the task creation here, since we would have to create a task in
                    # any branch of the _api_websocket_dispatch() to asynchronously do writing to websockets then.
                    asyncio.create_task(self._api_websocket_dispatch(request, ws, msg))
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    logger.info('API websocket connection closed with exception %s', ws.exception())
        finally:
            logger.debug('API websocket connection closed')
            self._websockets.discard(ws)
            for api_object in self._api_objects.values():
                api_object.websocket_close(ws)
            if ws in self._api_ws_last_will:
                obj, value = self._api_ws_last_will.pop(ws)
                await obj.http_post(value, ws)
            return ws

    async def _api_websocket_dispatch(self, request: aiohttp.web.Request, ws: aiohttp.web.WebSocketResponse,
                                      msg: aiohttp.WSMessage) -> None:
        try:
            message = msg.json()
        except JSONDecodeError:
            logger.warning("Websocket API message from %s is not a valid JSON string: %s", request.remote, msg.data)
            await ws.send_json({'status': 400, 'error': "Could not parse message as JSON: {}".format(msg.data)})
            return

        try:
            name = message["name"]
            action = message["action"]
            handle = message.get("handle")
        except KeyError:
            logger.warning("Websocket API message from %s without 'name' or 'action' field: %s", request.remote,
                           message)
            await ws.send_json({'status': 422,
                                'error': "Message does not include a 'name' and an 'action' field"})
            return
        result = {'status': 204,
                  'name': name,
                  'action': action,
                  'handle': handle}
        try:
            obj = self._api_objects[name]
        except KeyError:
            logger.warning("Could not find API object %s, requested by %s", name, request.remote)
            result['status'] = 404
            result['error'] = "There is no API object with name '{}'".format(name)
            await ws.send_json(result)
            return

        try:
            # subscribe action
            if action == "subscribe":
                logger.debug("got websocket subscribe request for API object %s from %s", name, request.remote)
                await obj.websocket_subscribe(ws, handle)
                return

            # post action
            elif action == "post":
                value_exists = False
                try:
                    value = message["value"]
                    value_exists = True
                except KeyError:
                    result['status'] = 422
                    result['error'] = "message does not include a 'value' field"
                    logger.warning("Websocket API POST message from %s without 'value' field: %s", request.remote,
                                   message)
                if value_exists:
                    logger.debug("got post request for API object %s via websocket from %s with value %s",
                                 name, request.remote, value)
                    try:
                        await obj.http_post(value, ws)
                    except (ValueError, TypeError) as e:
                        logger.warning("Error while updating API object %s with value via websocket from %s (error was "
                                       "%s): %s", name, request.remote, e, value)
                        result['status'] = 422
                        result['error'] = "Could not use provided value to update API object: {}".format(e)

            # lastwill action
            elif action == "lastwill":
                value_exists = False
                try:
                    value = message["value"]
                    value_exists = True
                except KeyError:
                    result['status'] = 422
                    result['error'] = "message does not include a 'value' field"
                    logger.warning("Websocket API LASTWILL message from %s without 'value' field: %s", request.remote,
                                   message)
                if value_exists:
                    logger.debug("got LASTWILL request for API object %s via websocket from %s with value %s",
                                 name, request.remote, value)
                    try:
                        self._api_ws_last_will[ws] = (obj, obj._check_last_will(value))
                    except (ValueError, TypeError) as e:
                        logger.warning("Error while setting last will of websocket client %s for API object %s (error"
                                       "was %s): %s", name, request.remote, e, value)
                        result['status'] = 422
                        result['error'] = "Could not use provided value to set last will: {}".format(e)

            # get action
            elif action == "get":
                logger.debug("got get request for API object %s via websocket from %s", name, request.remote)
                value = (await obj.http_get())[1]
                result['status'] = 200 if value is not None else 409
                result['value'] = value

            else:
                logger.warning("Unknown websocket API action '%s', requested by %s", action, request.remote)
                result['status'] = 422
                result['error'] = "Not a valid action: '{}'".format(action)
        except Exception as e:
            logger.error("Error while processing API websocket message from %s: %s", request.remote, message,
                         exc_info=e)
            result['status'] = 500
            result['error'] = "Internal server error while processing message"

        # Finally, send a response
        logger.debug("Sending websocket response: %s", result)
        await ws.send_str(json.dumps(result, cls=SHCJsonEncoder))

    async def _api_get_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        try:
            api_object = self._api_objects[request.match_info['name']]
        except KeyError:
            name_failsafe = request.match_info.get('name', '<undefined>')
            logger.warning("Could not find API object %s, requested by %s", name_failsafe, request.remote)
            raise aiohttp.web.HTTPNotFound(reason="Could not find API Object with name {}"
                                           .format(name_failsafe))
        # Parse `wait` and `timeout` from request query string
        wait = 'wait' in request.query
        timeout = 30.0
        if wait and request.query['wait']:
            try:
                timeout = float(request.query['wait'])
            except ValueError as e:
                raise aiohttp.web.HTTPBadRequest(reason="Could not parse 'wait' query parameter's value as float: {}"
                                                 .format(e))

        # if `wait`: Make this Request gracefully stoppable on shutdown by registering it for
        if wait:
            current_task = asyncio.current_task()
            assert(current_task is not None)
            self._associated_tasks.add(current_task)

        # Now, let's actually call http_get of the API object
        # If `wait`, this will await a new value or the `timeout`.
        changed, value, etag = await api_object.http_get(wait, timeout, request.headers.get('If-None-Match'))

        # If not changed (either when `wait` and timeout is reached) or if not `wait` and `If-None-Match` indicates
        # unchanged value, return HTTP 304 Not Modified
        if not changed:
            raise aiohttp.web.HTTPNotModified(headers={'ETag': etag})
        else:
            return aiohttp.web.Response(status=200 if value is not None else 409,
                                        headers={'ETag': etag},
                                        body=json.dumps(value, cls=SHCJsonEncoder),
                                        content_type="application/json",
                                        charset='utf-8')

    async def _api_post_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        text = await request.text()
        try:
            data = json.loads(text)
        except JSONDecodeError as e:
            logger.warning("Invalid JSON body POSTed from %s to %s (error was: %s): %s",
                           request.remote, request.url, e, text)
            raise aiohttp.web.HTTPBadRequest(reason="Could not parse request body as json: {}".format(str(e)))

        try:
            name = request.match_info['name']
            api_object = self._api_objects[name]
        except KeyError:
            name_failsafe = request.match_info.get('name', '<undefined>')
            logger.warning("Could not find API object %s, requested by %s", name_failsafe, request.remote)
            raise aiohttp.web.HTTPNotFound(reason="Could not find API Object with name {}"
                                           .format(name_failsafe))
        try:
            await api_object.http_post(data, request)
        except (ValueError, TypeError) as e:
            logger.warning("Error while updating API object %s with value from %s (error was %s): %s", name,
                           request.remote, e, data)
            raise aiohttp.web.HTTPUnprocessableEntity(reason="Could not use provided value to update API object: {}"
                                                      .format(e))
        raise aiohttp.web.HTTPNoContent()

    async def _monitoring_handler(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        accept_map = {'text/*': 'text/plain',
                      'application/*': 'application/json',
                      '*/*': 'application/json',
                      'text/html': 'text/html',
                      'application/json': 'application/json',
                      }
        accept_list = request.headers.get("Accept", "application/json").split(",")
        content_type: Optional[str] = None
        for mime_type in accept_list:
            mime_type = mime_type.strip().split(";")[0]
            if mime_type in accept_map:
                content_type = accept_map[mime_type]
                break
        if content_type is None:
            raise aiohttp.web.HTTPNotAcceptable()

        # Fetch interface data
        interfaces_data = {}
        overall_status = 0
        for iface in get_interfaces():
            status = await iface.get_status()
            interfaces_data[repr(iface)] = {
                'status': status.status.value,
                'message': status.message,
                'indicators': status.indicators,
            }
            overall_status = max(overall_status, min(status.status.value, 2) - 2 + iface.criticality.value)

        # Calculate HTTP status code
        http_status = {0: 200,
                       1: 213,
                       2: 513}.get(overall_status, 500)

        if content_type == "application/json":
            data = {
                'status': overall_status,
                'interfaces': interfaces_data,
            }
            body = json.dumps(data)
        elif content_type == "text/html":
            template = jinja_env.get_template('status.htm')
            body = await template.render_async(overall_status=overall_status, interfaces_data=interfaces_data,
                                               ServiceStatus=ServiceStatus, menu=self.ui_menu_entries,
                                               root_url=self.root_url, js_files=self._js_files,
                                               css_files=self._css_files, server_token=id(self),
                                               html_title="Status Monitoring")
        return aiohttp.web.Response(status=http_status,
                                    body=body,
                                    content_type=content_type,
                                    charset='utf-8')

    def serve_static_file(self, path: pathlib.Path) -> str:
        """
        Register a static file to be served on this HTTP server.

        The URL is automatically chosen, based on the file's name and existing static files.
        If the same path has already been added as a static file, its existing static URL is returned instead of
        creating a new one.

        This method should primarily be used by WebPageItem implementations within their
        :meth:`WebPageItem.register_with_server` method. It is meant for serving configuration-specific images etc.

        :param path: The path of the local file to be served as a static file
        :return: The URL of the static file, including the server's root_url, such that it is represented an absolute
            URL or absolute-path reference (relative to the HTTP server root), which can be used in <img>, <link> tags,
            etc.
        """
        path = path.absolute()
        if path in self.static_files:
            return f"{self.root_url}/addon/{self.static_files[path]}"

        final_file_name = path.name
        i = 0
        while final_file_name in self.static_files.values():
            final_file_name = "{}_{:04d}.{}".format(path.stem, i, path.suffix)
        self.static_files[path] = final_file_name
        final_url = '/addon/{}'.format(final_file_name)

        # Unfortunately, aiohttp.web.static can only serve directories. We want to serve a single file here.
        async def send_file(_request):
            return aiohttp.web.FileResponse(path)
        self._app.add_routes([aiohttp.web.get(final_url, send_file)])

        return self.root_url + final_url

    def add_static_directory(self, path: pathlib.Path, js_files: Iterable[str] = (), css_files: Iterable[str] = ()
                             ) -> str:
        """
        Register an additional directory of static files served by this server, optionally with a list of JavaScript and
        CSS files to be loaded in each page's HTML head.

        This method adds the given path as a static directory to be served by this webserver. The root URL of the served
        directory is automatically determined, based on the file's name and existing static files. If the same path has
        already been added as a static file, its existing static URL is returned instead of creating a new one.

        The given `js_files` and `css_files` are interpreted as relative URL references of files within the directory,
        that will be served by the HTTP server at the directory's root URL + this path. Make sure to use '/' as path
        separator and URL-encode these strings if necessary.

        :param path: Local filesystem path of directory to be served
        :param js_files: list of relative URLs within the directory to be included as Javascript files into UI pages
        :param css_files: list of relative URLs within the directory to be included as CSS files into UI pages
        :return: The HTTP root URL where the directory is served, including the server's root_url, such that it is
            represented an absolute URL or absolute-path reference (relative to the HTTP server root)
        """
        path = path.absolute()
        if path in self.static_files:
            final_url = '/addon/{}'.format(self.static_files[path])
        else:
            final_file_name = path.name
            i = 0
            while final_file_name in self.static_files.values():
                final_file_name = "{}_{:04d}.{}".format(path.stem, i, path.suffix)

            self.static_files[path] = final_file_name
            final_url = '/addon/{}'.format(final_file_name)

            self._app.add_routes([aiohttp.web.static(final_url, path)])

        self._css_files.extend((f"{final_url}/{file}" for file in css_files))
        self._js_files.extend((f"{final_url}/{file}" for file in js_files))

        return f"{self.root_url}{final_url}/"

    def __repr__(self) -> str:
        return "{}(host={}, port={})".format(self.__class__.__name__, self.host, self.port)


class WebConnectorContainer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_connectors(self) -> Iterable["WebUIConnector"]:
        pass


class WebPage(WebConnectorContainer):
    """
    Programmatic representation of a web UI page.

    To create a new page or get an existing page by name, use the :meth:`WebServer.page` method of the target web
    interface.
    """
    def __init__(self, server: WebServer, name: str, title: str):
        self.server = server
        self.name = name
        self.title = title
        self.segments: List["_WebPageSegment"] = []

    def add_item(self, item: "WebPageItem") -> None:
        """
        Add a new `WebPageItem` (widget) to the page.

        The item is appended to the current segment (see :meth:`new_segment`). If the page does not have any segments
        yet, a first segment is created with the default parameters (half-width, no heading) to contain the item.

        :param item: The `WebPageItem` to be added to this page
        """
        if not self.segments:
            self.new_segment()
        self.segments[-1].items.append(item)
        item.register_with_server(self, self.server)

    def get_connectors(self) -> Iterable["WebUIConnector"]:
        return itertools.chain.from_iterable(item.get_connectors() for item in self.segments)

    def new_segment(self, title: Optional[str] = None, same_column: bool = False, full_width: bool = False) -> None:
        """
        Create a new visual segment on the page, to contain further `WebPageItems`, optionally with a heading.

        :param title: A title for the segment, which is shown as a heading above the segment
        :param same_column: If True, the segment is added below the previous segment, in the same column (left or
            right). Otherwise, by default, the other column is used, which creates a new row, if the previous segment
            is in the right column. This option has no effect, when used for the first segment of a page.
        :param full_width: If True, the segment spans the full width of the page layout on large screens (1127px at
            maximum), instead of using only one of the two columns. In this case, the `same_column` parameter has no
            effect.
        """
        self.segments.append(_WebPageSegment(title, same_column, full_width))


class _WebPageSegment(WebConnectorContainer):
    def __init__(self, title: Optional[str], same_column: bool, full_width: bool):
        self.title = title
        self.same_column = same_column
        self.full_width = full_width
        self.items: List[WebPageItem] = []

    def get_connectors(self) -> Iterable[Union["WebUIConnector"]]:
        return itertools.chain.from_iterable(item.get_connectors() for item in self.items)


class WebPageItem(WebConnectorContainer, metaclass=abc.ABCMeta):
    """
    Abstract base class for all web UI widgets which can be added to a web UI page.
    """
    def register_with_server(self, page: WebPage, server: WebServer) -> None:
        """
        Called when the WebPageItem is added to a WebPage.

        It may be overidden by inheriting classes to get certain information about the WebPage or the WebServer or
        register required static files with the WebServer, using :meth:`WebServer.serve_static_file`,
        :meth:`WebServer.add_js_file`, :meth:`WebServer.serve_static_file`.

        :param page: The WebPage, this WebPageItem is added to.
        :param server: The WebServer, the WebPage (and thus, from now on, this WebPageItem) belongs to.
        """
        pass

    @abc.abstractmethod
    async def render(self) -> str:
        """
        Generate the HTML code of this `WebPageItem`.

        This coroutine is called as part of the rendering of the `WebPage`, the `WebPageItem` has been added to. It must
        be overriden by inheriting classes to return HTML code of the specific widget class to be inserted into the
        pages HTML code. To create interactive widgets, the HTML code should contain tags with `data-widget` and
        `data-id` attributes.

        :return: The HTML code of the specific `WebPageItem`
        """
        pass


class WebUIConnector(WebConnectorContainer, metaclass=abc.ABCMeta):
    """
    An abstract base class for all objects that want to exchange messages with JavaScript UI Widgets via the websocket
    connection.

    For every Message received from a client websocket, the :meth:`from_websocket` method of the appropriate
    `WebUIConnector` is called. For this purpose, the :class:`WebServer` creates a dict of all WebConnectors in any
    registered :class:`WebPage` by their Python object id at startup. The message from the websocket is expected to have
    an `id` field which is used for the lookup.
    """
    def __init__(self):
        self.subscribed_websockets: Set[aiohttp.web.WebSocketResponse] = set()

    def from_websocket(self, value: Any, ws: aiohttp.web.WebSocketResponse) -> None:
        """
        This method is called for incoming "value" messages from a client to this specific `WebUIConnector` object.

        It should be overridden by concrete `WebUIConnector` implementations to handle incoming values.

        :param value: The JSON-decoded 'value' field from the message from the websocket
        :param ws: The websocket connection, the message has been received from.
        """
        pass

    async def websocket_subscribe(self, ws: aiohttp.web.WebSocketResponse) -> None:
        await self._websocket_before_subscribe(ws)
        self.subscribed_websockets.add(ws)

    async def _websocket_before_subscribe(self, ws: aiohttp.web.WebSocketResponse) -> None:
        """
        This method is called by :meth:`websocket_subscribe`, when a new websocket subscribes to this specific
        `WebUIConnector`, *before* the client is added to the `subscribed_websockets` variable.

        It can be overridden to send an initial value or other initialization data to the client.
        """
        pass

    async def _websocket_publish(self, value: Any) -> None:
        """
        Send a value to the to all websocket clients subscribed to this `WebUIConnector` object

        This will trigger a call to the `update()` method of the subscribed JavaScript Widget objects.

        :param value: The value to send to the clients. Must be JSON-serializable using the `SHCJsonEncoder`, i.e. it
            should only include standard JSON types or types which have been registered for JSON conversion via
            :func:`shc.conversion.register_json_conversion`.
        """
        logger.debug("Publishing value %s for %s for %s subscribed websockets ...",
                     value, id(self), len(self.subscribed_websockets))
        data = json.dumps({'id': id(self), 'v': value}, cls=SHCJsonEncoder)
        await asyncio.gather(*(ws.send_str(data) for ws in self.subscribed_websockets))

    def websocket_close(self, ws: aiohttp.web.WebSocketResponse) -> None:
        self.subscribed_websockets.discard(ws)

    def get_connectors(self) -> Iterable["WebUIConnector"]:
        return (self,)

    def __repr__(self):
        return "{}<id={}>".format(self.__class__.__name__, id(self))


class WebDisplayDatapoint(Reading[T], Writable[T], WebUIConnector, metaclass=abc.ABCMeta):
    """
    Abstract base class for `WebUIConnectors` for state-displaying web UI widgets.

    This base class inherits from :class:`WebUIConnector` as well as the *Connectable* base classes :class:`Writable`
    and :class:`Reading`, which allows to *read* and *receive* values from another *Connectable* and forward them over
    the websocket to update a UI widget. This way, widgets reflecting the current value of a *Connectable* object can be
    built.

    This base class may be mixed with :class:`WebActionDatapoint`, creating a `Writable` + `Subscribable` class, to
    build interactive Widgets which display **and** update the connected objects' value.

    As this is a generic *Connectable* class, don't forget to define the :ref:`type attribute <base.typing>`, when
    inheriting from it—either as a class attribute or as an instance attribute, set in the constructor.
    """
    is_reading_optional = False

    async def _write(self, value: T, origin: List[Any]):
        if isinstance(self, WebActionDatapoint):
            self._publish(value, origin)
        await self._websocket_publish(self.convert_to_ws_value(value))

    def convert_to_ws_value(self, value: T) -> Any:
        """
        Callback method to convert new (*received* or *read*) values, before being JSON-encoded and published to the
        websocket clients

        This method may be overridden by inheriting classes to do any transformation of the new value, including type
        conversions. For example, a complex value of the object's *value type* may be used to evaluate a logic
        expression and only send the boolean result to the UI widget.

        Defaults to the identity function (simply returning the new value as is).

        :param value: The new value, as *read* or *received* from another *Connectable* object
        :return: The value to JSON-encoded and published to all subscribed websocket clients. Must be JSON-serializable
            using the `SHCJsonEncoder`, i.e. it should only include standard JSON types or types which have been
            registered for JSON conversion via :func:`shc.conversion.register_json_conversion`.
        """
        return value

    async def _websocket_before_subscribe(self, ws: aiohttp.web.WebSocketResponse) -> None:
        if self._default_provider is None:
            logger.error("Cannot handle websocket subscription for %s, since not read provider is registered.",
                         self)
            return
        logger.debug("New websocket subscription for widget id %s.", id(self))
        self.subscribed_websockets.add(ws)
        current_value = await self._from_provider()
        if current_value is not None:
            data = json.dumps({'id': id(self),
                               'v': self.convert_to_ws_value(current_value)},
                              cls=SHCJsonEncoder)
            await ws.send_str(data)


class WebActionDatapoint(Subscribable[T], WebUIConnector, metaclass=abc.ABCMeta):
    """
    Abstract base class for `WebUIConnectors` for interactive web UI widgets, publishing values to other objects.

    This base class inherits from :class:`WebUIConnector` as well as the *Connectable* base class :class:`Subscribable`,
    which allows it to *publish* values to other *Connectable* objects, when the UI widget sends a websocket message.
    This way, interactive widgets can be built, which publish values when the user interacts with them.

    This base class may be mixed with :class:`WebDisplayDatapoint`, creating a `Writable` + `Subscribable` class, to
    build interactive Widgets which display **and** update the connected objects' value.

    As this is a generic *Connectable* class, don't forget to define the :ref:`type attribute <base.typing>`, when
    inheriting from it—either as a class attribute or as an instance attribute, set in the constructor.
    """
    def __init__(self):
        super().__init__()
        self._stateful_publishing = isinstance(self, WebDisplayDatapoint)

    def convert_from_ws_value(self, value: Any) -> T:
        """
        Callback method to convert/transform values from a websocket client, before *publishing* them.

        This method may be overridden by inheriting classes to do any transformation of the new value, including type
        conversions. For example, a None-value may be transformed to a static value of the object's *value type*.

        Defaults to the identity function (simply returning the value as is).

        :param value: The JSON-decoded value, received from the websocket client
        :return: The value to be *published* to subscribed *Connectable* objects.
        """
        return from_json(self.type, value)

    def from_websocket(self, value: Any, ws: aiohttp.web.WebSocketResponse) -> None:
        value_converted = self.convert_from_ws_value(value)
        self._publish(value_converted, [ws])
        if isinstance(self, WebDisplayDatapoint):
            # from_websocket is awaited in the websocket message processing loop, so we should do all (asynchronously)
            # blocking things in separate tasks.
            asyncio.create_task(self._websocket_publish(self.convert_to_ws_value(value_converted)))


class WebApiObject(Reading[T], Writable[T], Subscribable[T], Generic[T]):
    """
    *Connectable* object that represents an endpoint of the REST/websocket API.

    :ivar name: The name of this object in the REST/websocket API
    """
    is_reading_optional = False
    _stateful_publishing = True

    def __init__(self, type_: Type[T], name: str):
        self.type = type_
        super().__init__()
        self.name = name
        self.subscribed_websockets: Set[aiohttp.web.WebSocketResponse] = set()
        self.future: asyncio.Future[T]

    def start(self) -> None:
        """
        Do some things at startup of the webserver.
        """
        # We do this upon server startup to ensure that the future is bound to the correct AsyncIO event loop.
        # This might not be the case if we create the future in the __init__ method, since the object creation might be
        # done in a different Thread than the SHC main event loop thread (as in our unittests).
        self.future = asyncio.get_event_loop().create_future()

    async def _write(self, value: T, origin: List[Any]) -> None:
        # Asynchronous local feedback publishing. This ensures that conflicting updates, which are currently waiting for
        # local processing by a subscriber can be corrected by resetting this update's origin.
        self._publish(value, origin)
        await self._publish_http(value)

    async def http_post(self, value: Any, origin: Any) -> None:
        self._publish(from_json(self.type, value), [origin])
        await self._publish_http(value, origin)

    def _check_last_will(self, value: Any) -> T:
        """
        Check and convert a value that has been send by an API client as a last will for this object.

        We do this synchronously when the client set its last will, so we can inform it if the value is invalid.

        :param value: The decoded json value from the client
        :return: The converted value to be stored for later use as a last will
        """
        return from_json(self.type, value)

    async def _post_last_will(self, value: T, ws: aiohttp.web.WebSocketResponse) -> None:
        """
        Post the stored last will of a client
        """
        self._publish(from_json(self.type, value), [ws])
        await self._publish_http(value, ws)

    async def _publish_http(self, value: T, skip_websocket: Optional[object] = None) -> None:
        """
        Publish a new value to all subscribed websockets and waiting long-running poll requests.

        :param skip_websocket: If given and a websocket object (:class:`aiohttp.web.WebSocketResponse`) which is
            subscribed to this WebApiObject, this websocket is skipped when publishing the new value. This should be
            used by :meth:`http_post` to prevent reflecting every value posted via the websocket API to its original
            sender.
        """
        self.future.set_result(value)
        self.future = asyncio.get_event_loop().create_future()
        data = json.dumps({'status': 200, 'name': self.name, 'value': value}, cls=SHCJsonEncoder)
        await asyncio.gather(*(ws.send_str(data) for ws in self.subscribed_websockets
                               if ws is not skip_websocket))

    async def websocket_subscribe(self, ws: aiohttp.web.WebSocketResponse, handle: Any) -> None:
        self.subscribed_websockets.add(ws)
        current_value = await self._from_provider()
        if current_value is not None:
            data = {'status': 200, 'action': 'subscribe', 'name': self.name, 'value': current_value, 'handle': handle}
        else:
            data = {'status': 204, 'action': 'subscribe', 'name': self.name, 'handle': handle}
        await ws.send_str(json.dumps(data, cls=SHCJsonEncoder))

    def websocket_close(self, ws: aiohttp.web.WebSocketResponse) -> None:
        self.subscribed_websockets.discard(ws)

    async def http_get(self, wait: bool = False, timeout: float = 30, etag_match: Optional[str] = None
                       ) -> Tuple[bool, Any, str]:
        """
        Get the current value or await a new value.

        This method is used for normal GET requests and long-running polls that only return when a new value is
        awailable.

        :param wait: If True, the method awaits the receiving of a new value or the expiration of the timeout. If False,
            it simply *reads* and returns the current value
        :param timeout: If `wait` is True and no value arrives within `timeout` seconds, this method returns, with the
            first element in the result tuple set to True to indicate the timeout.
        :param etag_match: The `If-None-Match` header value provided by the client. Should be the `etag` from the
            client's last call to this method.
            With `wait=True`: If given and not equal to the id of the current future, this method assumes that the
            client missed a value and falls back to *read* and return the current value immediately. This way, we make
            sure that the client does not miss an update while renewing its poll request.
            With `wait=False`: Normal HTTP behaviour: If the etag does match the current future's is, we return
            with `changed=False`, which should result in an
        :return: A tuple (changed, value, etag).
            `changed` is False, if this method returns due to a timeout or with `wait=False` and an etag indicating an
            unchanged value, or True, if due to a new value. Should be used for the HTTP status code: 200 vs. 304.
            `value` represents the new value (or None when `changed=False`).
            `etag` is the id of the new future. It can be used as the HTML `ETag` header, so the client can send it in
            the `If-None-Match` header of the next request, which is passed to this method's `etag_match` parameter.
        """
        # If not waiting for next value and etag indicates unchanged value: return with `changed=False`
        if not wait and etag_match == str(id(self.future)):
            return False, None, str(id(self.future))
        # If not waiting for next value *or* etag indicates changed value: return current value
        if not wait or (etag_match is not None and etag_match != str(id(self.future))):
            value = await self._from_provider()
            return True, value, str(id(self.future))

        # If waiting for next value: Await future using timeout
        try:
            value = await asyncio.wait_for(asyncio.shield(self.future), timeout=timeout)
            return True, value, str(id(self.future))
        except (asyncio.TimeoutError, asyncio.CancelledError):
            return False, None, str(id(self.future))



import asyncio
import enum
import json
import math
import shutil
import socket
import time
import unittest
import unittest.mock
import urllib.request
import urllib.error
import http.client
from pathlib import Path

import aiohttp
from selenium import webdriver
import selenium.webdriver.firefox.options
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver import ActionChains

import shc.web
import shc.web.widgets
from shc.datatypes import RangeFloat1, RGBUInt8, RangeUInt8
from shc.supervisor import AbstractInterface, InterfaceStatus, ServiceStatus
from ._helper import InterfaceThreadRunner, ExampleReadable, AsyncMock, async_test


class StatusTestInterface(AbstractInterface):
    def __init__(self, name):
        super().__init__()
        self.status = InterfaceStatus()
        self.name = name

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass

    async def get_status(self) -> "InterfaceStatus":
        return self.status

    def __repr__(self):
        return f"StatusTestInterface({self.name})"


@unittest.skipIf(shutil.which("geckodriver") is None, "Selenium's geckodriver is not available in PATH")
class AbstractWebTest(unittest.TestCase):
    driver: webdriver.Firefox

    def setUp(self) -> None:
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
        self.server = self.server_runner.interface

    @classmethod
    def setUpClass(cls) -> None:
        opts = selenium.webdriver.firefox.options.Options()
        opts.add_argument("-headless")
        cls.driver = webdriver.Firefox(options=opts)

    def tearDown(self) -> None:
        self.server_runner.stop()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.driver.close()
        cls.driver.quit()


class SimpleWebTest(AbstractWebTest):
    def test_basic(self) -> None:
        self.server_runner.start()
        self.driver.get("http://localhost:42080")

    def test_page(self) -> None:
        page = self.server.page('index', 'Home Page')
        page.add_item(shc.web.widgets.ButtonGroup("My button group", [
            shc.web.widgets.StatelessButton(42, "Foobar")
        ]))
        page.new_segment("Another segment", full_width=True)
        page.add_item(shc.web.widgets.ButtonGroup("Another button group", [
            shc.web.widgets.StatelessButton(42, "Bar")
        ]))

        self.server_runner.start()
        self.driver.get("http://localhost:42080")

        self.assertIn('Home Page', self.driver.page_source)
        self.assertIn('Home Page', self.driver.title)
        self.assertIn('Another segment', self.driver.page_source)
        button = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "Foobar"]')
        self.assertIn("My button group", button.find_element(By.XPATH, '../..').text)
        button = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "Bar"]')
        self.assertIn("Another button group", button.find_element(By.XPATH, '../..').text)

    def test_main_menu(self) -> None:
        self.server.page('index', menu_entry="Home", menu_icon='home')
        self.server.add_menu_entry('another_page', label="Foo", sub_label="Bar", sub_icon="bars")

        self.server_runner.start()
        self.driver.get("http://localhost:42080")

        # Only search for the (visible) main navigation bar, instead of the hidden sidebare for mobile screens:
        container = self.driver.find_element(By.CSS_SELECTOR, '.pusher')

        home_link = container.find_element(By.CSS_SELECTOR, 'i.home.icon').find_element(By.XPATH, '..')
        self.assertIn("Home", home_link.text)
        home_link.click()

        container = self.driver.find_element(By.CSS_SELECTOR, '.pusher')
        submenu = container.find_element(By.XPATH, './/div[contains(text(), "Foo")]')
        submenu_entry = submenu.find_element(By.XPATH, './/a[contains(@class, "item")]')
        self.assertFalse(submenu_entry.is_displayed())
        submenu.click()
        self.assertIn("Bar", submenu_entry.text)
        self.assertTrue(submenu_entry.is_displayed())
        self.assertEqual(submenu_entry.get_attribute('href').strip(), "http://localhost:42080/page/another_page/")
        submenu_entry.find_element(By.CSS_SELECTOR, 'i.bars.icon')


class MonitoringTest(unittest.TestCase):
    def setUp(self) -> None:
        shc.supervisor._REGISTERED_INTERFACES.clear()
        self.interface1 = StatusTestInterface("Interface 1")
        self.interface1.criticality = shc.supervisor.ServiceCriticality.INFO
        self.interface2 = StatusTestInterface("Interface 2")
        self.interface2.criticality = shc.supervisor.ServiceCriticality.WARNING
        self.interface3 = StatusTestInterface("Interface 3")
        self.interface3.criticality = shc.supervisor.ServiceCriticality.CRITICAL
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
        self.server = self.server_runner.interface

    def tearDown(self) -> None:
        self.server_runner.stop()
        shc.supervisor._REGISTERED_INTERFACES.clear()

    @async_test
    async def test_monitoring_json(self) -> None:
        self.server_runner.start()

        self.interface1.status = InterfaceStatus(ServiceStatus.CRITICAL, "Something is wrong", {"badness": 100})
        self.interface3.status = InterfaceStatus(ServiceStatus.WARNING, "Be warned",
                                                 {"info": "intervention may be required"})
        headers = {"Accept": "application/json"}
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.get('http://localhost:42080/monitoring') as resp:
                self.assertEqual(213, resp.status)
                data = await resp.json()

            self.assertEqual(1, data['status'])
            self.assertEqual(2, data['interfaces']['StatusTestInterface(Interface 1)']['status'])
            self.assertEqual("Something is wrong", data['interfaces']['StatusTestInterface(Interface 1)']['message'])
            self.assertEqual(100, data['interfaces']['StatusTestInterface(Interface 1)']['indicators']['badness'])
            self.assertEqual(0, data['interfaces']['StatusTestInterface(Interface 2)']['status'])
            self.assertEqual("", data['interfaces']['StatusTestInterface(Interface 2)']['message'])
            self.assertEqual(1, data['interfaces']['StatusTestInterface(Interface 3)']['status'])

            self.interface3.status = InterfaceStatus(ServiceStatus.CRITICAL, "ERROR ERROR!!1!")

            async with session.get('http://localhost:42080/monitoring') as resp:
                self.assertEqual(513, resp.status)
                data = await resp.json()
                self.assertEqual(2, data['status'])

            self.interface3.status = InterfaceStatus(ServiceStatus.OK, "Service restored")

            async with session.get('http://localhost:42080/monitoring') as resp:
                self.assertEqual(200, resp.status)
                data = await resp.json()
                self.assertEqual(0, data['status'])

    # TODO test HTML/UI monitoring page


class WebWidgetsTest(AbstractWebTest):
    def test_switch(self) -> None:
        page = self.server.page('index')
        switch_widget = shc.web.widgets.Switch("Main Power").connect(ExampleReadable(bool, True))
        page.add_item(switch_widget)

        with unittest.mock.patch.object(switch_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            checkbox_element = self.driver.find_element(
                By.XPATH, '//*[normalize-space(text()) = "Main Power"]/..//input')
            self.assertTrue(checkbox_element.is_selected())

            self.server_runner.run_coro(switch_widget.write(False, [self]))
            time.sleep(0.05)
            publish_mock.reset_mock()
            self.assertFalse(checkbox_element.is_selected())

            checkbox_container = checkbox_element.find_element(By.XPATH, './..')
            checkbox_container.click()
            time.sleep(0.05)
            self.assertTrue(checkbox_element.is_selected())
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)

    def test_switch_confirm(self) -> None:
        switch = shc.web.widgets.Switch("Some Switch", confirm_values=(True,), confirm_message="My text")\
            .connect(ExampleReadable(bool, False))

        page = self.server.page('index')
        page.add_item(switch)

        with unittest.mock.patch.object(switch, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            checkbox_container = self.driver.find_element(
                By.XPATH, '//*[normalize-space(text()) = "Some Switch"]/..//input/..')

            # Setting to true requires confirmation
            checkbox_container.click()
            time.sleep(0.05)
            publish_mock.assert_not_called()
            alert = Alert(self.driver)
            self.assertEqual("My text", alert.text)
            alert.accept()
            time.sleep(0.1)
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)

            # Setting back to false should not require a confirmation
            publish_mock.reset_mock()
            checkbox_container.click()
            time.sleep(0.05)
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)

    def test_buttons(self) -> None:
        b1 = shc.web.widgets.ToggleButton(label="B1", color='yellow')
        b2: shc.web.widgets.DisplayButton[bool] = shc.web.widgets.DisplayButton(label="B2", color='blue')
        b3 = shc.web.widgets.StatelessButton(42, "B3")
        b4 = shc.web.widgets.ValueButton(42, "B4", color="red")
        ExampleReadable(bool, True).connect(b1)
        ExampleReadable(bool, True).connect(b2)
        ExampleReadable(int, 42).connect(b4)

        page = self.server.page('index')
        page.add_item(shc.web.widgets.ButtonGroup("My button group", [b1, b2, b3, b4]))

        with unittest.mock.patch.object(b1, '_publish') as b1_publish,\
                unittest.mock.patch.object(b3, '_publish') as b3_publish,\
                unittest.mock.patch.object(b4, '_publish') as b4_publish:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)

            b1_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')
            b2_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B2"]')
            b3_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B3"]')
            b4_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B4"]')

            # Check initial states
            self.assertIn('yellow', b1_element.get_attribute('class'))
            self.assertIn('blue', b2_element.get_attribute('class'))
            self.assertIn('red', b4_element.get_attribute('class'))

            # Check state updates
            self.server_runner.run_coro(b1.write(False, [self]))
            time.sleep(0.05)
            self.assertNotIn('yellow', b1_element.get_attribute('class'))
            self.assertIn('blue', b2_element.get_attribute('class'))
            self.assertIn('red', b4_element.get_attribute('class'))

            self.server_runner.run_coro(b2.write(False, [self]))
            self.server_runner.run_coro(b4.write(56, [self]))
            time.sleep(0.05)
            b1_publish.reset_mock()
            b3_publish.reset_mock()
            b4_publish.reset_mock()
            self.assertNotIn('blue', b2_element.get_attribute('class'))
            self.assertNotIn('red', b4_element.get_attribute('class'))

            # Check clicks
            b1_element.click()
            time.sleep(0.05)
            b1_publish.assert_called_once_with(True, unittest.mock.ANY)
            b3_publish.assert_not_called()
            b4_publish.assert_not_called()

            b3_element.click()
            time.sleep(0.05)
            b1_publish.assert_called_once()
            b3_publish.assert_called_once_with(42, unittest.mock.ANY)
            b4_publish.assert_not_called()

            b4_element.click()
            time.sleep(0.05)
            b1_publish.assert_called_once()
            b3_publish.assert_called_once()
            b4_publish.assert_called_once_with(42, unittest.mock.ANY)

    def test_button_confirm(self) -> None:
        button = shc.web.widgets.ToggleButton(label="B1", color='yellow', confirm_message="Sure?",
                                              confirm_values=(True,))

        page = self.server.page('index')
        page.add_item(shc.web.widgets.ButtonGroup("My button group", [button]))

        with unittest.mock.patch.object(button, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)

            button_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')

            # Check click with alert
            self.assertNotIn('yellow', button_element.get_attribute('class'))
            button_element.click()
            time.sleep(0.05)
            publish_mock.assert_not_called()

            # dismiss alert
            alert = Alert(self.driver)
            self.assertEqual("Sure?", alert.text)
            alert.dismiss()
            time.sleep(0.05)
            publish_mock.assert_not_called()
            self.assertNotIn('yellow', button_element.get_attribute('class'))

            # Click again and accept alert
            button_element.click()
            time.sleep(0.01)
            alert = Alert(self.driver)
            alert.accept()
            time.sleep(0.05)
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)
            self.assertIn('yellow', button_element.get_attribute('class'))

            # No alert for setting to False
            publish_mock.reset_mock()
            button_element.click()
            time.sleep(0.05)
            publish_mock.assert_called_once_with(False, unittest.mock.ANY)
            self.assertNotIn('yellow', button_element.get_attribute('class'))

    def test_display(self) -> None:
        page = self.server.page('index')
        text_widget = shc.web.widgets.TextDisplay(int, "{} lux", "Brightness").connect(ExampleReadable(int, 42))
        page.add_item(text_widget)

        self.server_runner.start()
        self.driver.get("http://localhost:42080")
        time.sleep(0.4)
        value_element = self.driver.find_element(
            By.XPATH, '//*[normalize-space(text()) = "Brightness"]/..//*[@data-id]')
        self.assertEqual("42 lux", value_element.text.strip())

        self.server_runner.run_coro(text_widget.write(56, [self]))
        time.sleep(0.05)
        self.assertEqual("56 lux", value_element.text.strip())

    def test_input_int(self) -> None:
        page = self.server.page('index')
        input_widget = shc.web.widgets.TextInput(int, "Brightness").connect(ExampleReadable(int, 42))
        page.add_item(input_widget)

        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            input_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Brightness"]/..//input')
            self.assertEqual("42", input_element.get_attribute("value"))

            self.server_runner.run_coro(input_widget.write(56, [self]))
            time.sleep(0.05)
            publish_mock.reset_mock()
            self.assertEqual("56", input_element.get_attribute("value"))

            input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
            input_element.send_keys("15", Keys.ENTER)
            time.sleep(0.05)
            self.assertEqual("15", input_element.get_attribute("value"))
            publish_mock.assert_called_once_with(15, unittest.mock.ANY)

            input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
            input_element.send_keys("18", Keys.ESCAPE)
            time.sleep(0.05)
            self.assertEqual("15", input_element.get_attribute("value"))
            publish_mock.assert_called_once()

            input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
            input_element.send_keys("42", Keys.TAB)
            time.sleep(0.05)
            self.assertEqual("42", input_element.get_attribute("value"))
            publish_mock.assert_called_with(42, unittest.mock.ANY)

    def test_input_string(self) -> None:
        page = self.server.page('index')
        input_widget = shc.web.widgets.TextInput(str, "Message of the Day")\
            .connect(ExampleReadable(str, "Hello, World!"))
        page.add_item(input_widget)

        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            input_element = self.driver.find_element(
                By.XPATH, '//*[normalize-space(text()) = "Message of the Day"]/..//input')
            self.assertEqual("Hello, World!", input_element.get_attribute("value"))

            self.server_runner.run_coro(input_widget.write("Foobar", [self]))
            time.sleep(0.05)
            publish_mock.reset_mock()
            self.assertEqual("Foobar", input_element.get_attribute("value"))

            input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
            input_element.send_keys("Hello, SHC!", Keys.ENTER)
            time.sleep(0.05)
            self.assertEqual("Hello, SHC!", input_element.get_attribute("value"))
            publish_mock.assert_called_once_with("Hello, SHC!", unittest.mock.ANY)

    def test_slider(self) -> None:
        page = self.server.page('index')
        input_widget = shc.web.widgets.Slider("Amount of Foo").connect(ExampleReadable(RangeFloat1, RangeFloat1(0.3)))
        page.add_item(input_widget)

        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            container_element = self.driver.find_element(
                By.XPATH, '//*[normalize-space(text()) = "Amount of Foo"]/../..')
            slider_element = container_element.find_element(By.CSS_SELECTOR, '.slider')
            handle_element = slider_element.find_element(By.CSS_SELECTOR, ".thumb")

            # Center of handle should be somewhere near 30% of width of
            slider_width = slider_element.rect['width']
            slider_start = slider_element.rect['x']
            handle_center = handle_element.rect['x'] + handle_element.rect['width']/2
            self.assertAlmostEqual(slider_start + 0.3 * slider_width, handle_center, delta=4)  # 4px off is okay

            # Let's move the handle to about 70%
            ActionChains(self.driver).drag_and_drop_by_offset(handle_element, 0.4 * slider_width, 0).perform()

            time.sleep(0.05)
            publish_mock.assert_called_once()
            self.assertAlmostEqual(0.7, publish_mock.call_args[0][0], delta=0.05)  # 5% off is okay

            # Let's move the handle to about 0%
            publish_mock.reset_mock()
            ActionChains(self.driver).drag_and_drop_by_offset(handle_element, -0.8 * slider_width, 0).perform()
            time.sleep(0.05)
            publish_mock.assert_called_once()
            self.assertEqual(0.0, publish_mock.call_args[0][0])

    def test_hiderow(self) -> None:
        page = self.server.page('index')
        foo_button = shc.web.widgets.StatelessButton(True, shc.web.widgets.icon('power off'))
        foo_row = shc.web.widgets.HideRow("Foo", foo_button, 'red').connect(ExampleReadable(bool, True))
        bar_row = shc.web.widgets.HideRow("Bar").connect(ExampleReadable(bool, True))
        foobar_row = shc.web.widgets.HideRow("Foobar", color='yellow').connect(ExampleReadable(bool, False))
        page.add_item(shc.web.widgets.HideRowBox([foo_row, bar_row, foobar_row]))

        with unittest.mock.patch.object(foo_button, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            foo_row_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Foo"]')
            bar_row_element = self.driver.find_element(By.XPATH, '//*[normalize-space(text()) = "Bar"]')
            # Foobar row should not be findable, since Selenium's xpath can only find DOM elements by text content if
            # they are visible
            foobar_row_element = self.driver.find_element(By.XPATH, '//*[contains(text(), "Foobar")]')
            button = foo_row_element.find_element(By.CSS_SELECTOR, "button")

            self.assertTrue(foo_row_element.is_displayed())
            self.assertTrue(bar_row_element.is_displayed())
            self.assertFalse(foobar_row_element.is_displayed())
            self.assertIn('red', foo_row_element.get_attribute('class'))
            self.assertIn('blue', bar_row_element.get_attribute('class'))
            self.assertIn('yellow', foobar_row_element.get_attribute('class'))
            self.assertIn('power off', button.find_element(By.CSS_SELECTOR, '.icon').get_attribute('class'))

            # Click the button
            button.click()
            time.sleep(0.05)
            publish_mock.assert_called_once_with(True, unittest.mock.ANY)

            # Swap the rows
            self.server_runner.run_coro(foo_row.write(False, [self]))
            self.server_runner.run_coro(bar_row.write(False, [self]))
            self.server_runner.run_coro(foobar_row.write(True, [self]))
            time.sleep(1)
            self.assertFalse(foo_row_element.is_displayed())
            self.assertFalse(bar_row_element.is_displayed())
            self.assertTrue(foobar_row_element.is_displayed())

    def test_colorchoser(self) -> None:
        page = self.server.page('index')
        input_widget = shc.web.widgets.ColorChoser()\
            .connect(ExampleReadable(RGBUInt8, RGBUInt8(RangeUInt8(127), RangeUInt8(127), RangeUInt8(127))))
        page.add_item(input_widget)

        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            wheel_element = self.driver.find_element(By.CSS_SELECTOR, '.IroWheel')
            wheel_handle_element = wheel_element.find_element(By.CSS_SELECTOR, '.IroHandle>circle')
            slider_element = self.driver.find_element(By.CSS_SELECTOR, '.IroSlider')
            slider_handle_element = slider_element.find_element(By.CSS_SELECTOR, '.IroHandle>circle')

            # The wheel handle should be in the center of the wheel ...
            wheel_rect = wheel_element.rect
            wheel_center = (wheel_rect['x'] + wheel_rect['width']/2, wheel_rect['y'] + wheel_rect['height']/2)
            wheel_handle_rect = wheel_element.rect
            wheel_handle_center = (wheel_handle_rect['x'] + wheel_handle_rect['width']/2,
                                   wheel_handle_rect['y'] + wheel_handle_rect['height']/2)
            self.assertAlmostEqual(wheel_center[0], wheel_handle_center[0], delta=4)  # 4px off is okay
            self.assertAlmostEqual(wheel_center[1], wheel_handle_center[1], delta=4)  # 4px off is okay

            # The slider handle should be at 50%
            slider_rect = slider_element.rect
            slider_handle_rect = slider_handle_element.rect
            self.assertAlmostEqual(slider_rect['x'] + slider_rect['width']/2,
                                   slider_handle_rect['x'] + slider_handle_rect['width']/2, delta=4)  # 4px off is okay

            # Now, lets set a yellow color at 80% brightness
            # Yellow is in the right lower corner of the wheel, at 120° clockwise from the top or -60° (-pi/3 rad)
            # mathematically. (Attention: The y axis is inverted in contrast to the normal mathetmatical orientation)
            ActionChains(self.driver)\
                .move_to_element_with_offset(slider_element, 0.8 * slider_rect['width'], slider_rect['height']/2)\
                .click()\
                .move_to_element(wheel_handle_element)\
                .click_and_hold()\
                .move_to_element_with_offset(
                    wheel_element,
                    wheel_rect['width']/2 + 0.6 * math.cos(-math.pi/3) * wheel_rect['width'],
                    wheel_rect['height']/2 + -0.6 * math.sin(-math.pi/3) * wheel_rect['height'])\
                .release()\
                .perform()

            time.sleep(0.05)
            self.assertEqual(2, publish_mock.call_count)
            latest_color = publish_mock.call_args[0][0]
            self.assertAlmostEqual(204, latest_color.red, delta=13)  # 5% off is okay
            self.assertAlmostEqual(0, latest_color.green, delta=6)
            self.assertAlmostEqual(204, latest_color.blue, delta=13)

    def test_enum_select(self) -> None:
        class ExampleEnum(enum.Enum):
            SOME_VALUE = 0
            SOME_OTHER_VALUE = 1
            YET_ANOTHER_VALUE = 2

        page = self.server.page('index')
        input_widget = shc.web.widgets.EnumSelect(ExampleEnum, "Select the Foo")\
            .connect(ExampleReadable(ExampleEnum, ExampleEnum.SOME_OTHER_VALUE))
        page.add_item(input_widget)

        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
            self.server_runner.start()
            self.driver.get("http://localhost:42080")
            time.sleep(0.4)
            container_element = self.driver.find_element(
                By.XPATH, '//*[normalize-space(text()) = "Select the Foo"]/..')
            menu_element = container_element.find_element(By.CSS_SELECTOR, '.selection.dropdown')

            menu_element.click()
            second_option_element = menu_element\
                .find_element(By.XPATH, './/*[normalize-space(text()) = "SOME_OTHER_VALUE"][contains(@class, "item")]')
            self.assertIn("selected", second_option_element.get_attribute('class'))

            # select the third option
            third_option_element = menu_element\
                .find_element(By.XPATH, './/*[normalize-space(text()) = "YET_ANOTHER_VALUE"]')
            third_option_element.click()

            time.sleep(0.05)
            menu_element.click()
            self.assertNotIn("selected", second_option_element.get_attribute('class'))
            self.assertIn("selected", third_option_element.get_attribute('class'))
            publish_mock.assert_called_once_with(ExampleEnum.YET_ANOTHER_VALUE, unittest.mock.ANY)

    def test_image_map(self) -> None:
        page = self.server.page('index')
        b1 = shc.web.widgets.ToggleButton(label="B1", color='yellow').connect(ExampleReadable(bool, True))
        l2 = shc.web.widgets.ImageMapLabel(float, color='red').connect(ExampleReadable(float, 15.3))

        page.add_item(shc.web.widgets.ImageMap((Path(__file__)).parent / 'assets' / 'example_image.jpg', [
            (0.3, 0.3, b1),
            (0.9, 0.55, l2)
        ]))

        self.server_runner.start()
        self.driver.get("http://localhost:42080")
        time.sleep(0.6)

        # Check that the background image is served and loaded correctly
        background_image = self.driver.find_element(By.CSS_SELECTOR, '.shc.image-container .background')
        self.assertTrue(self.driver.execute_script(
            "return arguments[0].complete "
            "&& typeof arguments[0].naturalWidth != \"undefined\" "
            "&& arguments[0].naturalWidth == 600", background_image))
        background_image_rect = background_image.rect

        # Check that there is a correctly styled button element, labeled "B1", at the right position
        b1_element = self.driver.find_element(By.XPATH, '//button[normalize-space(text()) = "B1"]')
        self.assertIn('yellow', b1_element.get_attribute('class'))
        b1_rect = b1_element.rect
        self.assertAlmostEqual(background_image_rect['x'] + background_image_rect['width'] * 0.3,
                               b1_rect['x'] + b1_rect['width']/2, delta=4)  # 4px off is okay
        self.assertAlmostEqual(background_image_rect['y'] + background_image_rect['height'] * 0.3,
                               b1_rect['y'] + b1_rect['height']/2, delta=4)  # 4px off is okay

        # Check that there is a label in the right position
        l2_element = self.driver.find_element(By.CSS_SELECTOR, '.shc.image-container .ui.label')
        self.assertEqual("15.3", l2_element.text.strip())
        self.assertIn('red', l2_element.get_attribute('class'))
        l2_rect = l2_element.rect
        self.assertAlmostEqual(background_image_rect['x'] + background_image_rect['width'] * 0.9,
                               l2_rect['x'] + l2_rect['width']/2, delta=4)  # 4px off is okay
        self.assertAlmostEqual(background_image_rect['y'] + background_image_rect['height'] * 0.55,
                               l2_rect['y'] + l2_rect['height']/2, delta=4)  # 4px off is okay


class TestAPI(unittest.TestCase):
    # We use the Python built-in (synchronous) HTTP client (urllib.request / http.client) to test the compatibility with
    # another HTTP implementation and have a more realistic control flow/timing (with different threads instead of one
    # AsyncIO event loop)
    def setUp(self) -> None:
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
        self.server = self.server_runner.interface

    def tearDown(self) -> None:
        self.server_runner.stop()

    def test_rest_get(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()

        with self.assertRaises(urllib.error.HTTPError) as cm:
            urllib.request.urlopen("http://localhost:42080/api/v1/object/a_non_existing_object")
        self.assertEqual(404, cm.exception.code)

        response: http.client.HTTPResponse = urllib.request.urlopen(
            "http://localhost:42080/api/v1/object/the_api_object")
        self.assertEqual(42, json.loads(response.read()))
        etag1 = response.headers["ETag"]

        # GET request with 'If-None-Match' header with the retrieved ETag should return HTTP 304 NotModified
        request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object",
                                         headers={'If-None-Match': etag1})
        with self.assertRaises(urllib.error.HTTPError) as cm:
            urllib.request.urlopen(request)
        self.assertEqual(304, cm.exception.code)
        self.assertEqual(etag1, cm.exception.headers["ETag"])

    def test_rest_get_wait(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()

        async def scheduled_update(value):
            await asyncio.sleep(0.2)
            await api_object.write(value, [self])

        tic = time.time()
        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
        asyncio.run_coroutine_threadsafe(scheduled_update(56), self.server_runner.loop)

        response: http.client.HTTPResponse = urllib.request.urlopen(
            "http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
        toc = time.time()
        self.assertEqual(56, json.loads(response.read()))
        self.assertAlmostEqual(0.2, toc-tic, delta=0.02)

        # When no update happens, we should retrieve a HTTP 304 NotModified after 0.5 s
        tic = time.time()
        with self.assertRaises(urllib.error.HTTPError) as cm:
            urllib.request.urlopen("http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
        toc = time.time()
        self.assertEqual(304, cm.exception.getcode())
        self.assertAlmostEqual(0.5, toc-tic, delta=0.02)

    def test_rest_get_wait_etag(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()

        # A normal GET request to get the current ETag
        response: http.client.HTTPResponse = urllib.request.urlopen(
            "http://localhost:42080/api/v1/object/the_api_object")
        etag1 = response.headers["ETag"]

        async def scheduled_update(value):
            await asyncio.sleep(0.2)
            await api_object.write(value, [self])

        # A GET request with matching ETag and wait parameter. It should wait for the next value
        tic = time.time()
        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
        asyncio.run_coroutine_threadsafe(scheduled_update(56), self.server_runner.loop)

        request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object?wait=0.5",
                                         headers={'If-None-Match': etag1})
        response = urllib.request.urlopen(request)
        toc = time.time()
        etag2 = response.headers['ETag']
        self.assertEqual(56, json.loads(response.read()))
        self.assertAlmostEqual(0.2, toc-tic, delta=0.02)
        self.assertNotEqual(etag1, etag2)

        # Now, simulate that we missed that update by sending the same ETag again. It should return immediately with a
        # freshly read value (which is still 42) but the new ETag
        tic = time.time()
        request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object?wait=0.5",
                                         headers={'If-None-Match': etag1})
        response = urllib.request.urlopen(request)
        toc = time.time()
        self.assertEqual(42, json.loads(response.read()))
        self.assertEqual(etag2, response.headers['ETag'])
        self.assertAlmostEqual(0, toc-tic, delta=0.02)

    def test_rest_get_wait_abort(self) -> None:
        """ Test, that an aborted long poll to the API does not destroy any internal state """
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()

        async def scheduled_update(value):
            await asyncio.sleep(0.2)
            await api_object.write(value, [self])

        request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
        with self.assertRaises(socket.timeout):
            urllib.request.urlopen(request, timeout=0.25)

        # Check that everything still works:
        self.server_runner.run_coro(scheduled_update(56))

        tic = time.time()
        # The `scheduled_update()` coroutine shall run in parallel with our request below. Thus, we cannot use
        # InterfaceThreadRunner.run_coro(), which waits for the coroutine to return.
        asyncio.run_coroutine_threadsafe(scheduled_update(128), self.server_runner.loop)

        response: http.client.HTTPResponse = urllib.request.urlopen(
            "http://localhost:42080/api/v1/object/the_api_object?wait=0.5")
        toc = time.time()
        self.assertEqual(128, json.loads(response.read()))
        self.assertAlmostEqual(0.2, toc-tic, delta=0.02)

    def test_rest_post(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()

        # POST to non-existing object
        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
            with self.assertRaises(urllib.error.HTTPError) as cm:
                request = urllib.request.Request("http://localhost:42080/api/v1/object/non_existing_object",
                                                 data=json.dumps(56).encode(), method="POST")
                urllib.request.urlopen(request)
        self.assertEqual(404, cm.exception.code)
        publish_mock.assert_not_called()

        # valid POST
        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
            request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object",
                                             data=json.dumps(56).encode(), method="POST")
            urllib.request.urlopen(request)
        publish_mock.assert_called_once_with(56, unittest.mock.ANY)

        # POST with invalid JSON data → HTTP 400 InvalidRequest
        with self.assertRaises(urllib.error.HTTPError) as cm:
            request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object",
                                             data="56,".encode(), method="POST")
            urllib.request.urlopen(request)
        self.assertEqual(400, cm.exception.code)

        # POST with invalid JSON data type → HTTP 422 InvalidRequest
        with self.assertRaises(urllib.error.HTTPError) as cm:
            request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object",
                                             data=json.dumps("abc").encode(), method="POST")
            urllib.request.urlopen(request)
        self.assertEqual(422, cm.exception.code)


class WebSocketAPITest(unittest.TestCase):
    def setUp(self) -> None:
        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
        self.server = self.server_runner.interface

        self.closing = False
        self.ws_callback = unittest.mock.Mock()

    async def start_websocket(self) -> None:
        async def handle_websocket(ws):
            async for msg in ws:
                if msg.type in (aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY):
                    self.ws_callback(msg.data)
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    if self.closing:
                        break
                    else:
                        raise AssertionError("Websocket closed by server.")
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    raise AssertionError("Websocket error: {}".format(msg.data))

        self.client_session = aiohttp.ClientSession()
        self.ws = await self.client_session.ws_connect('http://localhost:42080/api/v1/ws')
        self.ws_receiver_task = asyncio.create_task(handle_websocket(self.ws))

    def tearDown(self) -> None:
        # Close client
        # websocket
        self.closing = True
        loop = asyncio.get_event_loop()
        loop.create_task(self.ws.close())
        loop.create_task(self.client_session.close())
        pending = asyncio.all_tasks(loop)
        loop.run_until_complete(asyncio.gather(*pending))
        # Await ws receiver task to catch websocket errors
        loop.run_until_complete(self.ws_receiver_task)

        self.server_runner.stop()

    @async_test
    async def test_errors(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()
        await self.start_websocket()

        # Invalid JSON → 400
        await self.ws.send_str("42,")
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual(400, data['status'])
        self.assertIn('JSON', data['error'])

        # Missing action → 422
        self.ws_callback.reset_mock()
        await self.ws.send_json({'name': 'the_api_object'})
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual(422, data['status'])
        self.assertIn('action', data['error'])

        # Missing object name → 422
        self.ws_callback.reset_mock()
        await self.ws.send_json({'action': 'get'})
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual(422, data['status'])
        self.assertIn('name', data['error'])

        # Invalid action → 422
        self.ws_callback.reset_mock()
        await self.ws.send_json({'action': 'foobar', 'name': 'the_api_object'})
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual(422, data['status'])
        self.assertIn('action', data['error'])

        # Unknown object → 404
        self.ws_callback.reset_mock()
        await self.ws.send_json({'action': 'get', 'name': 'non_existing_object'})
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual(404, data['status'])
        self.assertIn('name', data['error'])

    @async_test
    async def test_get(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        self.server_runner.start()
        await self.start_websocket()

        await self.ws.send_json({'action': 'get', 'name': 'the_api_object'})
        await asyncio.sleep(0.05)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual('get', data['action'])
        self.assertEqual('the_api_object', data['name'])
        self.assertEqual(200, data['status'])
        self.assertEqual(42, data['value'])

    @async_test
    async def test_post(self) -> None:
        api_object = self.server.api(int, "the_api_object")
        self.server_runner.start()
        await self.start_websocket()

        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
            await self.ws.send_json({'action': 'post', 'name': 'the_api_object', 'value': 56})
            await asyncio.sleep(0.05)

        publish_mock.assert_called_once_with(56, unittest.mock.ANY)
        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual('post', data['action'])
        self.assertEqual('the_api_object', data['name'])
        self.assertEqual(204, data['status'])

    @async_test
    async def test_subscribe(self) -> None:
        api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
        api_object2 = self.server.api(int, "the_other_api_object")
        self.server_runner.start()
        await self.start_websocket()

        await self.ws.send_json({'action': 'subscribe', 'name': 'the_api_object'})
        await asyncio.sleep(0.05)

        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual('subscribe', data['action'])
        self.assertEqual('the_api_object', data['name'])
        self.assertEqual(42, data['value'])
        self.assertEqual(200, data['status'])

        self.ws_callback.reset_mock()
        await self.ws.send_json({'action': 'subscribe', 'name': 'the_other_api_object', 'handle': 42})
        await asyncio.sleep(0.05)

        self.ws_callback.assert_called_once()
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertEqual('subscribe', data['action'])
        self.assertEqual('the_other_api_object', data['name'])
        self.assertEqual(42, data['handle'])
        self.assertNotIn('value', data)

        self.ws_callback.reset_mock()
        self.server_runner.run_coro(api_object.write(56, [self]))
        self.server_runner.run_coro(api_object2.write(56, [self]))
        await asyncio.sleep(0.05)

        self.assertEqual(2, self.ws_callback.call_count)
        data = json.loads(self.ws_callback.call_args[0][0])
        self.assertIn(data['name'], ('the_api_object', 'the_other_api_object'))
        self.assertEqual(56, data['value'])
        self.assertEqual(200, data['status'])

    @async_test
    async def test_last_will(self) -> None:
        api_object = self.server.api(int, "the_api_object")
        self.server_runner.start()
        await self.start_websocket()

        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
            # First lastwill request is broken (wrong value type)
            await self.ws.send_json({'action': 'lastwill', 'name': 'the_api_object', 'value': "hello, world!"})
            await asyncio.sleep(0.05)

            publish_mock.assert_not_called()

            self.ws_callback.assert_called_once()
            data = json.loads(self.ws_callback.call_args[0][0])
            self.assertEqual('lastwill', data['action'])
            self.assertEqual('the_api_object', data['name'])
            self.assertEqual(422, data['status'])
            self.ws_callback.reset_mock()

            # Now, a correct lastwill request. It should still not publish any value at the server
            await self.ws.send_json({'action': 'lastwill', 'name': 'the_api_object', 'value': 56})
            await asyncio.sleep(0.05)

            publish_mock.assert_not_called()

            self.ws_callback.assert_called_once()
            data = json.loads(self.ws_callback.call_args[0][0])
            self.assertEqual('lastwill', data['action'])
            self.assertEqual('the_api_object', data['name'])
            self.assertEqual(204, data['status'])

            # Now, let the client disconnect. This should trigger publishing the value at the server
            await self.ws.close()
            await self.client_session.close()
            await asyncio.sleep(0.1)

            publish_mock.assert_called_once_with(56, unittest.mock.ANY)
